# Copyright (C) 2006-2008 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
"""Python bindings for Subversion."""

__author__ = "Jelmer Vernooij <jelmer@jelmer.uk>"

from subvertpy import SubversionException, ERR_BAD_URL

from subvertpy import _ra
from subvertpy._ra import *  # noqa: F403,F401
from subvertpy import ra_svn  # noqa: F401

try:
    from urllib2 import splittype
except ImportError:
    from urllib.parse import splittype

url_handlers = {
    "svn": _ra.RemoteAccess,
    # "svn": ra_svn.Client,
    "svn+ssh": _ra.RemoteAccess,
    # "svn+ssh": ra_svn.Client,
    "http": _ra.RemoteAccess,
    "https": _ra.RemoteAccess,
    "file": _ra.RemoteAccess,
}


def RemoteAccess(url, *args, **kwargs):
    """Connect to a remote Subversion server

    :param url: URL to connect to
    :return: RemoteAccess object
    """
    if isinstance(url, bytes):
        url = url.decode("utf-8")
    (type, opaque) = splittype(url)
    if type not in url_handlers:
        raise SubversionException("Unknown URL type '%s'" % type, ERR_BAD_URL)
    return url_handlers[type](url, *args, **kwargs)



# Copyright (C) 2017 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Subversion subr library tests."""

import os
from unittest import TestCase

from subvertpy.subr import (
    uri_canonicalize,
    dirent_canonicalize,
    abspath,
)


class UriCanonicalizeTests(TestCase):
    def test_canonicalize(self):
        self.assertEqual(
            "https://www.example.com", uri_canonicalize("https://www.example.com/")
        )
        self.assertEqual(
            "https://www.example.com(bla)",
            uri_canonicalize("https://www.example.com(bla)"),
        )
        self.assertEqual(
            "https://www.example.com/(bla)",
            uri_canonicalize("https://www.example.com/(bla%29"),
        )


class DirentCanonicalizeTests(TestCase):
    def test_canonicalize(self):
        self.assertEqual("/foo/bar", dirent_canonicalize("/foo/bar"))
        self.assertEqual("/foo/bar", dirent_canonicalize("/foo//bar"))


class AbspathTests(TestCase):
    def test_abspath(self):
        path = "/foo/bar" if os.name != "nt" else "C:/foo/bar"
        self.assertEqual(path, abspath(path))
        # os.getcwd() returns '/foo/bar' on linux/macos
        # while it returns 'c:\\foo\\bar' on windows
        self.assertEqual(
            os.path.join(os.getcwd(), "bar").replace("\\", "/").lower(),
            abspath("bar").lower(),
        )
        self.assertEqual(
            os.path.join(os.getcwd(), "bar", "foo").replace("\\", "/").lower(),
            abspath("bar/foo").lower(),
        )



# Copyright (C) 2006-2008 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
"""Python bindings for Subversion."""

__author__ = "Jelmer Vernooij <jelmer@jelmer.uk>"

try:
    from SocketServer import StreamRequestHandler, TCPServer
except ImportError:
    from socketserver import StreamRequestHandler, TCPServer
import base64
import os
import socket
import subprocess
from errno import EPIPE

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from subvertpy import (
    ERR_RA_SVN_UNKNOWN_CMD,
    ERR_UNSUPPORTED_FEATURE,
    NODE_DIR,
    NODE_FILE,
    NODE_UNKNOWN,
    NODE_NONE,
    SubversionException,
    properties,
)
from subvertpy.delta import (
    pack_svndiff0_window,
    unpack_svndiff0,
    SVNDIFF0_HEADER,
)
from subvertpy.marshall import (
    NeedMoreData,
    literal,
    marshall,
    unmarshall,
)
from subvertpy.ra import (
    DIRENT_CREATED_REV,
    DIRENT_HAS_PROPS,
    DIRENT_KIND,
    DIRENT_LAST_AUTHOR,
    DIRENT_SIZE,
    DIRENT_TIME,
)
from subvertpy.server import (
    generate_random_id,
)


class SSHSubprocess(object):
    """A socket-like object that talks to an ssh subprocess via pipes."""

    __slots__ = "proc"

    def __init__(self, proc):
        self.proc = proc

    def send(self, data):
        return os.write(self.proc.stdin.fileno(), data)

    def recv(self, count):
        return os.read(self.proc.stdout.fileno(), count)

    def close(self):
        self.proc.stdin.close()
        self.proc.stdout.close()
        self.proc.wait()

    def get_filelike_channels(self):
        return (self.proc.stdout, self.proc.stdin)


class SSHVendor(object):
    def connect_ssh(self, username, password, host, port, command):
        args = ["ssh", "-x"]
        if port is not None:
            args.extend(["-p", str(port)])
        if username is not None:
            host = "%s@%s" % (username, host)
        args.append(host)
        proc = subprocess.Popen(
            args + command, stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        return SSHSubprocess(proc)


# Can be overridden by users
get_ssh_vendor = SSHVendor


class SVNConnection(object):
    def __init__(self, recv_fn, send_fn):
        self.inbuffer = ""
        self.recv_fn = recv_fn
        self.send_fn = send_fn

    def recv_msg(self):
        while True:
            try:
                (self.inbuffer, ret) = unmarshall(self.inbuffer)
                return ret
            except NeedMoreData:
                newdata = self.recv_fn(1)
                if newdata != "":
                    # self.mutter("IN: %r" % newdata)
                    self.inbuffer += newdata

    def send_msg(self, data):
        marshalled_data = marshall(data)
        # self.mutter("OUT: %r" % marshalled_data)
        self.send_fn(marshalled_data)

    def send_success(self, *contents):
        self.send_msg([literal("success"), list(contents)])


SVN_PORT = 3690


def feed_editor(conn, editor):
    tokens = {}
    diff = {}
    txdelta_handler = {}
    # Process commands
    while True:
        command, args = conn.recv_msg()
        if command == "target-rev":
            editor.set_target_revision(args[0])
        elif command == "open-root":
            if len(args[0]) == 0:
                token = editor.open_root()
            else:
                token = editor.open_root(args[0][0])
            tokens[args[1]] = token
        elif command == "delete-entry":
            tokens[args[2]].delete_entry(args[0], args[1])
        elif command == "add-dir":
            if len(args[3]) == 0:
                token = tokens[args[1]].add_directory(args[0])
            else:
                token = tokens[args[1]].add_directory(args[0], args[3][0], args[4][0])
            tokens[args[2]] = token
        elif command == "open-dir":
            tokens[args[2]] = tokens[args[1]].open_directory(args[0], args[3])
        elif command == "change-dir-prop":
            if len(args[2]) == 0:
                tokens[args[0]].change_prop(args[1], None)
            else:
                tokens[args[0]].change_prop(args[1], args[2][0])
        elif command == "close-dir":
            tokens[args[0]].close()
        elif command == "absent-dir":
            tokens[args[1]].absent(args[0])
        elif command == "add-file":
            if len(args[3]) == 0:
                token = tokens[args[1]].add_file(args[0])
            else:
                token = tokens[args[1]].add_file(args[0], args[3][0], args[4][0])
            tokens[args[2]] = token
        elif command == "open-file":
            tokens[args[2]] = tokens[args[1]].open_file(args[0], args[3])
        elif command == "apply-textdelta":
            if len(args[1]) == 0:
                txdelta_handler[args[0]] = tokens[args[0]].apply_textdelta(None)
            else:
                txdelta_handler[args[0]] = tokens[args[0]].apply_textdelta(args[1][0])
            diff[args[0]] = ""
        elif command == "textdelta-chunk":
            diff[args[0]] += args[1]
        elif command == "textdelta-end":
            for w in unpack_svndiff0(diff[args[0]]):
                txdelta_handler[args[0]](w)
            txdelta_handler[args[0]](None)
        elif command == "change-file-prop":
            if len(args[2]) == 0:
                tokens[args[0]].change_prop(args[1], None)
            else:
                tokens[args[0]].change_prop(args[1], args[2][0])
        elif command == "close-file":
            if len(args[1]) == 0:
                tokens[args[0]].close()
            else:
                tokens[args[0]].close(args[1][0])
        elif command == "close-edit":
            editor.close()
            break
        elif command == "abort-edit":
            editor.abort()
            break

    conn.send_success()
    conn._unpack()


class Reporter(object):
    __slots__ = ("conn", "editor")

    def __init__(self, conn, editor):
        self.conn = conn
        self.editor = editor

    def set_path(self, path, rev, start_empty=False, lock_token=None, depth=None):
        args = [path, rev, start_empty]
        if lock_token is not None:
            args.append([lock_token])
        else:
            args.append([])
        if depth is not None:
            args.append(depth)

        self.conn.send_msg([literal("set-path"), args])

    def delete_path(self, path):
        self.conn.send_msg([literal("delete-path"), [path]])

    def link_path(self, path, url, rev, start_empty=False, lock_token=None, depth=None):
        args = [path, url, rev, start_empty]
        if lock_token is not None:
            args.append([lock_token])
        else:
            args.append([])
        if depth is not None:
            args.append(depth)

        self.conn.send_msg([literal("link-path"), args])

    def finish(self):
        self.conn.send_msg([literal("finish-report"), []])
        self.conn.recv_msg()
        feed_editor(self.conn, self.editor)
        self.conn.busy = False

    def abort(self):
        self.conn.send_msg([literal("abort-report"), []])
        self.conn.busy = False


class Editor(object):
    __slots__ = "conn"

    def __init__(self, conn):
        self.conn = conn

    def set_target_revision(self, revnum):
        self.conn.send_msg([literal("target-rev"), [revnum]])

    def open_root(self, base_revision=None):
        id = generate_random_id()
        if base_revision is None:
            baserev = []
        else:
            baserev = [base_revision]
        self.conn.send_msg([literal("open-root"), [baserev, id]])
        self.conn._open_ids = []
        return DirectoryEditor(self.conn, id)

    def close(self):
        self.conn.send_msg([literal("close-edit"), []])

    def abort(self):
        self.conn.send_msg([literal("abort-edit"), []])


class DirectoryEditor(object):
    __slots__ = ("conn", "id")

    def __init__(self, conn, id):
        self.conn = conn
        self.id = id
        self.conn._open_ids.append(id)

    def add_file(self, path, copyfrom_path=None, copyfrom_rev=-1):
        self._is_last_open()
        child = generate_random_id()
        if copyfrom_path is not None:
            copyfrom_data = [copyfrom_path, copyfrom_rev]
        else:
            copyfrom_data = []
        self.conn.send_msg([literal("add-file"), [path, self.id, child, copyfrom_data]])
        return FileEditor(self.conn, child)

    def open_file(self, path, base_revnum):
        self._is_last_open()
        child = generate_random_id()
        self.conn.send_msg([literal("open-file"), [path, self.id, child, base_revnum]])
        return FileEditor(self.conn, child)

    def delete_entry(self, path, base_revnum):
        self._is_last_open()
        self.conn.send_msg([literal("delete-entry"), [path, base_revnum, self.id]])

    def add_directory(self, path, copyfrom_path=None, copyfrom_rev=-1):
        self._is_last_open()
        child = generate_random_id()
        if copyfrom_path is not None:
            copyfrom_data = [copyfrom_path, copyfrom_rev]
        else:
            copyfrom_data = []
        self.conn.send_msg([literal("add-dir"), [path, self.id, child, copyfrom_data]])
        return DirectoryEditor(self.conn, child)

    def open_directory(self, path, base_revnum):
        self._is_last_open()
        child = generate_random_id()
        self.conn.send_msg([literal("open-dir"), [path, self.id, child, base_revnum]])
        return DirectoryEditor(self.conn, child)

    def change_prop(self, name, value):
        self._is_last_open()
        if value is None:
            value = []
        else:
            value = [value]
        self.conn.send_msg([literal("change-dir-prop"), [self.id, name, value]])

    def _is_last_open(self):
        assert self.conn._open_ids[-1] == self.id

    def close(self):
        self._is_last_open()
        self.conn._open_ids.pop()
        self.conn.send_msg([literal("close-dir"), [self.id]])


class FileEditor(object):
    __slots__ = ("conn", "id")

    def __init__(self, conn, id):
        self.conn = conn
        self.id = id
        self.conn._open_ids.append(id)

    def _is_last_open(self):
        assert self.conn._open_ids[-1] == self.id

    def close(self, checksum=None):
        self._is_last_open()
        self.conn._open_ids.pop()
        if checksum is None:
            checksum = []
        else:
            checksum = [checksum]
        self.conn.send_msg([literal("close-file"), [self.id, checksum]])

    def apply_textdelta(self, base_checksum=None):
        self._is_last_open()
        if base_checksum is None:
            base_check = []
        else:
            base_check = [base_checksum]
        self.conn.send_msg([literal("apply-textdelta"), [self.id, base_check]])
        self.conn.send_msg([literal("textdelta-chunk"), [self.id, SVNDIFF0_HEADER]])

        def send_textdelta(delta):
            if delta is None:
                self.conn.send_msg([literal("textdelta-end"), [self.id]])
            else:
                self.conn.send_msg(
                    [literal("textdelta-chunk"), [self.id, pack_svndiff0_window(delta)]]
                )

        return send_textdelta

    def change_prop(self, name, value):
        self._is_last_open()
        if value is None:
            value = []
        else:
            value = [value]
        self.conn.send_msg([literal("change-file-prop"), [self.id, name, value]])


def mark_busy(unbound):

    def convert(self, *args, **kwargs):
        self.busy = True
        try:
            ret = unbound(self, *args, **kwargs)
        finally:
            self.busy = False
        return ret

    convert.__doc__ = unbound.__doc__
    convert.__name__ = unbound.__name__
    return convert


def unmarshall_dirent(d):
    ret = {
        "name": d[0],
        "kind": d[1],
        "size": d[2],
        "has-props": bool(d[3]),
        "created-rev": d[4],
    }
    if d[5] != []:
        ret["created-date"] = d[5]
    if d[6] != []:
        ret["last-author"] = d[6]
    return ret


class SVNClient(SVNConnection):
    def __init__(
        self,
        url,
        progress_cb=None,
        auth=None,
        config=None,
        client_string_func=None,
        open_tmp_file_func=None,
    ):
        self.url = url
        (type, opaque) = urlparse.splittype(url)
        assert type in ("svn", "svn+ssh")
        (host, path) = urlparse.splithost(opaque)
        self._progress_cb = progress_cb
        self._auth = auth
        self._config = config
        self._client_string_func = client_string_func
        # open_tmp_file_func is ignored, as it is not needed for svn://
        if type == "svn":
            (recv_func, send_func) = self._connect(host)
        else:
            (recv_func, send_func) = self._connect_ssh(host)
        super(SVNClient, self).__init__(recv_func, send_func)
        (min_version, max_version, _, self._server_capabilities) = self._recv_greeting()
        self.send_msg(
            [
                max_version,
                [literal(x) for x in CAPABILITIES if x in self._server_capabilities],
                self.url,
            ]
        )
        (self._server_mechanisms, mech_arg) = self._unpack()
        if self._server_mechanisms != []:
            # FIXME: Support other mechanisms as well
            self.send_msg(
                [
                    literal("ANONYMOUS"),
                    [base64.b64encode("anonymous@%s" % socket.gethostname())],
                ]
            )
            self.recv_msg()
        msg = self._unpack()
        if len(msg) > 2:
            self._server_capabilities += msg[2]
        (self._uuid, self._root_url) = msg[0:2]
        self.busy = False

    def _unpack(self):
        msg = self.recv_msg()
        if msg[0] == "failure":
            if isinstance(msg[1], str):
                raise SubversionException(*msg[1])
            num = msg[1][0][0]
            msg = msg[1][0][1]
            if num == ERR_RA_SVN_UNKNOWN_CMD:
                raise NotImplementedError(msg)
            raise SubversionException(msg, num)
        assert msg[0] == "success", "Got: %r" % msg
        assert len(msg) == 2
        return msg[1]

    def _recv_greeting(self):
        greeting = self._unpack()
        assert len(greeting) == 4
        return greeting

    _recv_ack = _unpack

    def _connect(self, host):
        (host, port) = urlparse.splitnport(host, SVN_PORT)
        sockaddrs = socket.getaddrinfo(
            host, port, socket.AF_UNSPEC, socket.SOCK_STREAM, 0, 0
        )
        self._socket = None
        last_err = RuntimeError("no addresses for %s:%s" % (host, port))
        for family, socktype, proto, canonname, sockaddr in sockaddrs:
            try:
                self._socket = socket.socket(family, socktype, proto)
                self._socket.connect(sockaddr)
            except socket.error as err:
                last_err = err
                if self._socket is not None:
                    self._socket.close()
                self._socket = None
                continue
            break
        if self._socket is None:
            raise last_err
        self._socket.setblocking(True)
        return (self._socket.recv, self._socket.send)

    def _connect_ssh(self, host):
        (user, host) = urlparse.splituser(host)
        if user is not None:
            (user, password) = urlparse.splitpassword(user)
        else:
            password = None
        (host, port) = urlparse.splitnport(host, 22)
        self._tunnel = get_ssh_vendor().connect_ssh(
            user, password, host, port, ["svnserve", "-t"]
        )
        return (self._tunnel.recv, self._tunnel.send)

    def get_file_revs(self, path, start, end, file_rev_handler):
        raise NotImplementedError(self.get_file_revs)

    @mark_busy
    def get_locations(self, path, peg_revision, location_revisions):
        self.send_msg(
            [literal("get-locations"), [path, peg_revision, location_revisions]]
        )
        self._recv_ack()
        ret = {}
        while True:
            msg = self.recv_msg()
            if msg == "done":
                break
            ret[msg[0]] = msg[1]
        self._unparse()
        return ret

    def get_locks(self, path):
        self.send_msg([literal("get-lock"), [path]])
        self._recv_ack()
        return self._unpack()

    def lock(self, path_revs, comment, steal_lock, lock_func):
        raise NotImplementedError(self.lock)

    def unlock(self, path_tokens, break_lock, lock_func):
        raise NotImplementedError(self.unlock)

    def mergeinfo(self, paths, revision=-1, inherit=None, include_descendants=False):
        raise NotImplementedError(self.mergeinfo)

    def location_segments(
        self, path, start_revision, end_revision, include_merged_revisions=False
    ):
        args = [path]
        if start_revision is None or start_revision == -1:
            args.append([])
        else:
            args.append([start_revision])
        if end_revision is None or end_revision == -1:
            args.append([])
        else:
            args.append([end_revision])
        args.append(include_merged_revisions)
        self.send_msg([literal("get-location-segments"), args])
        self._recv_ack()
        while True:
            msg = self.recv_msg()
            if msg == "done":
                break
            yield msg
        self._unpack()

    def get_location_segments(self, path, start_revision, end_revision, rcvr):
        for msg in self.location_segments(path, start_revision, end_revision):
            rcvr(*msg)

    def has_capability(self, capability):
        return capability in self._server_capabilities

    @mark_busy
    def check_path(self, path, revision=None):
        args = [path]
        if revision is None or revision == -1:
            args.append([])
        else:
            args.append([revision])
        self.send_msg([literal("check-path"), args])
        self._recv_ack()
        ret = self._unpack()[0]
        return {
            "dir": NODE_DIR,
            "file": NODE_FILE,
            "unknown": NODE_UNKNOWN,
            "none": NODE_NONE,
        }[ret]

    def get_lock(self, path):
        self.send_msg([literal("get-lock"), [path]])
        self._recv_ack()
        ret = self._unpack()
        if len(ret) == 0:
            return None
        else:
            return ret[0]

    @mark_busy
    def get_dir(
        self, path, revision=-1, dirent_fields=0, want_props=True, want_contents=True
    ):
        args = [path]
        if revision is None or revision == -1:
            args.append([])
        else:
            args.append([revision])

        args += [want_props, want_contents]

        fields = []
        if dirent_fields & DIRENT_KIND:
            fields.append(literal("kind"))
        if dirent_fields & DIRENT_SIZE:
            fields.append(literal("size"))
        if dirent_fields & DIRENT_HAS_PROPS:
            fields.append(literal("has-props"))
        if dirent_fields & DIRENT_CREATED_REV:
            fields.append(literal("created-rev"))
        if dirent_fields & DIRENT_TIME:
            fields.append(literal("time"))
        if dirent_fields & DIRENT_LAST_AUTHOR:
            fields.append(literal("last-author"))
        args.append(fields)

        self.send_msg([literal("get-dir"), args])
        self._recv_ack()
        ret = self._unpack()
        fetch_rev = ret[0]
        props = dict(ret[1])
        dirents = {}
        for d in ret[2]:
            entry = unmarshall_dirent(d)
            dirents[entry["name"]] = entry

        return (dirents, fetch_rev, props)

    @mark_busy
    def stat(self, path, revision=-1):
        args = [path]
        if revision is None or revision == -1:
            args.append([revision])
        else:
            args.append([])

        self.send_msg([literal("stat"), args])
        self._recv_ack()
        ret = self._unpack()
        if len(ret) == 0:
            return None
        return unmarshall_dirent(ret[0])

    @mark_busy
    def get_file(self, path, stream, revision=-1):
        raise NotImplementedError(self.get_file)

    def change_rev_prop(self, rev, name, value):
        args = [rev, name]
        if value is not None:
            args.append(value)
        self.send_msg([literal("change-rev-prop"), args])
        self._recv_ack()
        self._unparse()

    def get_commit_editor(
        self, revprops, callback=None, lock_tokens=None, keep_locks=False
    ):
        args = [revprops[properties.PROP_REVISION_LOG]]
        if lock_tokens is not None:
            args.append(list(lock_tokens.items()))
        else:
            args.append([])
        args.append(keep_locks)
        if len(revprops) > 1:
            args.append(list(revprops.items()))
        self.send_msg([literal("commit"), args])
        self._recv_ack()
        raise NotImplementedError(self.get_commit_editor)

    def rev_proplist(self, revision):
        self.send_msg([literal("rev-proplist"), [revision]])
        self._recv_ack()
        return dict(self._unpack()[0])

    def rev_prop(self, revision, name):
        self.send_msg([literal("rev-prop"), [revision, name]])
        self._recv_ack()
        ret = self._unpack()
        if len(ret) == 0:
            return None
        else:
            return ret[0]

    @mark_busy
    def replay(self, revision, low_water_mark, update_editor, send_deltas=True):
        self.send_msg([literal("replay"), [revision, low_water_mark, send_deltas]])
        self._recv_ack()
        feed_editor(self, update_editor)
        self._unpack()

    @mark_busy
    def replay_range(
        self, start_revision, end_revision, low_water_mark, cbs, send_deltas=True
    ):
        self.send_msg(
            [
                literal("replay-range"),
                [start_revision, end_revision, low_water_mark, send_deltas],
            ]
        )
        self._recv_ack()
        for i in range(start_revision, end_revision + 1):
            msg = self.recv_msg()
            assert msg[0] == "revprops"
            edit = cbs[0](i, dict(msg[1]))
            feed_editor(self, edit)
            cbs[1](i, dict(msg[1]), edit)
        self._unpack()

    def do_switch(
        self,
        revision_to_update_to,
        update_target,
        recurse,
        switch_url,
        update_editor,
        depth=None,
    ):
        args = []
        if revision_to_update_to is None or revision_to_update_to == -1:
            args.append([])
        else:
            args.append([revision_to_update_to])
        args.append(update_target)
        args.append(recurse)
        args.append(switch_url)
        if depth is not None:
            args.append(literal(depth))

        self.busy = True
        try:
            self.send_msg([literal("switch"), args])
            self._recv_ack()
            return Reporter(self, update_editor)
        except BaseException:
            self.busy = False
            raise

    def do_update(
        self, revision_to_update_to, update_target, recurse, update_editor, depth=None
    ):
        args = []
        if revision_to_update_to is None or revision_to_update_to == -1:
            args.append([])
        else:
            args.append([revision_to_update_to])
        args.append(update_target)
        args.append(recurse)
        if depth is not None:
            args.append(literal(depth))

        self.busy = True
        try:
            self.send_msg([literal("update"), args])
            self._recv_ack()
            return Reporter(self, update_editor)
        except BaseException:
            self.busy = False
            raise

    def do_diff(
        self,
        revision_to_update,
        diff_target,
        versus_url,
        diff_editor,
        recurse=True,
        ignore_ancestry=False,
        text_deltas=False,
        depth=None,
    ):
        args = []
        if revision_to_update is None or revision_to_update == -1:
            args.append([])
        else:
            args.append([revision_to_update])
        args += [diff_target, recurse, ignore_ancestry, versus_url, text_deltas]
        if depth is not None:
            args.append(literal(depth))
        self.busy = True
        try:
            self.send_msg([literal("diff"), args])
            self._recv_ack()
            return Reporter(self, diff_editor)
        except BaseException:
            self.busy = False
            raise

    def get_repos_root(self):
        return self._root_url

    @mark_busy
    def get_latest_revnum(self):
        self.send_msg([literal("get-latest-rev"), []])
        self._recv_ack()
        return self._unpack()[0]

    @mark_busy
    def get_dated_rev(self, date):
        self.send_msg([literal("get-dated-rev"), [date]])
        self._recv_ack()
        return self._unpack()[0]

    @mark_busy
    def reparent(self, url):
        self.send_msg([literal("reparent"), [url]])
        self._recv_ack()
        self._unpack()
        self.url = url

    def get_uuid(self):
        return self._uuid

    @mark_busy
    def log(
        self,
        paths,
        start,
        end,
        limit=0,
        discover_changed_paths=True,
        strict_node_history=True,
        include_merged_revisions=True,
        revprops=None,
    ):
        args = [paths]
        if start is None or start == -1:
            args.append([])
        else:
            args.append([start])
        if end is None or end == -1:
            args.append([])
        else:
            args.append([end])
        args.append(discover_changed_paths)
        args.append(strict_node_history)
        args.append(limit)
        args.append(include_merged_revisions)
        if revprops is None:
            args.append(literal("all-revprops"))
            args.append([])
        else:
            args.append(literal("revprops"))
            args.append(revprops)

        self.send_msg([literal("log"), args])
        self._recv_ack()
        while True:
            msg = self.recv_msg()
            if msg == "done":
                break
            paths = {}
            for p, action, cfd in msg[0]:
                if len(cfd) == 0:
                    paths[p] = (str(action), None, -1)
                else:
                    paths[p] = (str(action), cfd[0], cfd[1])

            if len(msg) > 5:
                has_children = msg[5]
            else:
                has_children = None
            if len(msg) > 6 and msg[6]:
                revno = None
            else:
                revno = msg[1]  # noqa: F841
                # TODO(jelmer): Do something with revno
            revprops = {}
            if len(msg[2]) != 0:
                revprops[properties.PROP_REVISION_AUTHOR] = msg[2][0]
            if len(msg[3]) != 0:
                revprops[properties.PROP_REVISION_DATE] = msg[3][0]
            if len(msg[4]) != 0:
                revprops[properties.PROP_REVISION_LOG] = msg[4][0]
            if len(msg) > 8:
                revprops.update(dict(msg[8]))
            yield paths, msg[1], revprops, has_children

        self._unpack()

    def get_log(self, callback, *args, **kwargs):
        for paths, rev, props, has_children in self.log(*args, **kwargs):
            if has_children is None:
                callback(paths, rev, props)
            else:
                callback(paths, rev, props, has_children)


MIN_VERSION = 2
MAX_VERSION = 2
CAPABILITIES = ["edit-pipeline", "bazaar", "log-revprops"]
MECHANISMS = ["ANONYMOUS"]


class SVNServer(SVNConnection):
    def __init__(self, backend, recv_fn, send_fn, logf=None):
        self.backend = backend
        self._stop = False
        self._logf = logf
        super(SVNServer, self).__init__(recv_fn, send_fn)

        self.send_success(
            MIN_VERSION,
            MAX_VERSION,
            [literal(x) for x in MECHANISMS],
            [literal(x) for x in CAPABILITIES],
        )

    def send_mechs(self):
        self.send_success([literal(x) for x in MECHANISMS], "")

    def send_failure(self, *contents):
        self.send_msg([literal("failure"), list(contents)])

    def send_ack(self):
        self.send_success([], "")

    def send_unknown(self, cmd):
        self.send_failure(
            [ERR_RA_SVN_UNKNOWN_CMD, "Unknown command '%s'" % cmd, __file__, 52]
        )

    def get_latest_rev(self):
        self.send_ack()
        self.send_success(self.repo_backend.get_latest_revnum())

    def check_path(self, path, rev):
        if len(rev) == 0:
            revnum = None
        else:
            revnum = rev[0]
        kind = self.repo_backend.check_path(path, revnum)
        self.send_ack()
        self.send_success(
            literal(
                {
                    NODE_NONE: "none",
                    NODE_DIR: "dir",
                    NODE_FILE: "file",
                    NODE_UNKNOWN: "unknown",
                }[kind]
            )
        )

    def log(
        self,
        target_path,
        start_rev,
        end_rev,
        changed_paths,
        strict_node,
        limit=None,
        include_merged_revisions=False,
        all_revprops=None,
        revprops=None,
    ):
        def send_revision(revno, author, date, message, changed_paths=None):
            changes = []
            if changed_paths is not None:
                for p, (action, cf, cr) in changed_paths.items():
                    if cf is not None:
                        changes.append((p, literal(action), (cf, cr)))
                    else:
                        changes.append((p, literal(action), ()))
            self.send_msg([changes, revno, [author], [date], [message]])

        self.send_ack()
        if len(start_rev) == 0:
            start_revnum = None
        else:
            start_revnum = start_rev[0]
        if len(end_rev) == 0:
            end_revnum = None
        else:
            end_revnum = end_rev[0]
        self.repo_backend.log(
            send_revision,
            target_path,
            start_revnum,
            end_revnum,
            changed_paths,
            strict_node,
            limit,
        )
        self.send_msg(literal("done"))
        self.send_success()

    def open_backend(self, url):
        (rooturl, location) = urlparse.splithost(url)
        self.repo_backend, self.relpath = self.backend.open_repository(location)

    def reparent(self, parent):
        self.open_backend(parent)
        self.send_ack()
        self.send_success()

    def stat(self, path, rev):
        if len(rev) == 0:
            revnum = None
        else:
            revnum = rev[0]
        self.send_ack()
        dirent = self.repo_backend.stat(path, revnum)
        if dirent is None:
            self.send_success([])
        else:
            args = [
                dirent["name"],
                dirent["kind"],
                dirent["size"],
                dirent["has-props"],
                dirent["created-rev"],
            ]
            if "created-date" in dirent:
                args.append([dirent["created-date"]])
            else:
                args.append([])
            if "last-author" in dirent:
                args.append([dirent["last-author"]])
            else:
                args.append([])
            self.send_success([args])

    def commit(self, logmsg, locks, keep_locks=False, rev_props=None):
        self.send_failure(
            [ERR_UNSUPPORTED_FEATURE, "commit not yet supported", __file__, 42]
        )

    def rev_proplist(self, revnum):
        self.send_ack()
        revprops = self.repo_backend.rev_proplist(revnum)
        self.send_success(list(revprops.items()))

    def rev_prop(self, revnum, name):
        self.send_ack()
        revprops = self.repo_backend.rev_proplist(revnum)
        if name in revprops:
            self.send_success([revprops[name]])
        else:
            self.send_success()

    def get_locations(self, path, peg_revnum, revnums):
        self.send_ack()
        locations = self.repo_backend.get_locations(path, peg_revnum, revnums)
        for rev, path in locations.items():
            self.send_msg([rev, path])
        self.send_msg(literal("done"))
        self.send_success()

    def update(self, rev, target, recurse, depth=None, send_copyfrom_param=True):
        self.send_ack()
        while True:
            msg = self.recv_msg()
            assert msg[0] in ["set-path", "finish-report"]
            if msg[0] == "finish-report":
                break

        self.send_ack()

        if len(rev) == 0:
            revnum = None
        else:
            revnum = rev[0]
        self.repo_backend.update(Editor(self), revnum, target, recurse)
        self.send_success()
        client_result = self.recv_msg()
        if client_result[0] == "success":
            return
        else:
            self.mutter("Client reported error during update: %r" % client_result)
            # Needs to be sent back to the client to display
            self.send_failure(client_result[1][0])

    commands = {
        "get-latest-rev": get_latest_rev,
        "log": log,
        "update": update,
        "check-path": check_path,
        "reparent": reparent,
        "stat": stat,
        "commit": commit,
        "rev-proplist": rev_proplist,
        "rev-prop": rev_prop,
        "get-locations": get_locations,
        # FIXME: get-dated-rev
        # FIXME: get-file
        # FIXME: get-dir
        # FIXME: check-path
        # FIXME: switch
        # FIXME: status
        # FIXME: diff
        # FIXME: get-file-revs
        # FIXME: replay
    }

    def send_auth_request(self):
        pass

    def serve(self):
        self.send_greeting()
        msg = self.recv_msg()
        version = msg[0]
        capabilities = msg[1]
        url = msg[2]
        if len(msg) > 3:
            self.client_user_agent = msg[3]
        else:
            self.client_user_agent = None
        self.capabilities = capabilities
        self.version = version
        self.url = url
        self.mutter("client supports:")
        self.mutter("  version %r" % version)
        self.mutter("  capabilities %r " % capabilities)
        self.send_mechs()

        (mech, args) = self.recv_msg()
        # TODO: Proper authentication
        self.send_success()

        self.open_backend(url)
        self.send_success(self.repo_backend.get_uuid(), url)

        # Expect:
        while not self._stop:
            (cmd, args) = self.recv_msg()
            if cmd not in self.commands:
                self.mutter("client used unknown command %r" % cmd)
                self.send_unknown(cmd)
                return
            else:
                self.commands[cmd](self, *args)

    def close(self):
        self._stop = True

    def mutter(self, text):
        if self._logf is not None:
            self._logf.write("%s\n" % text)


class TCPSVNRequestHandler(StreamRequestHandler):
    def __init__(self, request, client_address, server):
        self._server = server
        StreamRequestHandler.__init__(self, request, client_address, server)

    def handle(self):
        server = SVNServer(
            self._server._backend, self.rfile.read, self.wfile.write, self._server._logf
        )
        try:
            server.serve()
        except socket.error as e:
            if e.args[0] == EPIPE:
                return
            raise


class TCPSVNServer(TCPServer):
    allow_reuse_address = True
    serve = TCPServer.serve_forever

    def __init__(self, backend, addr, logf=None):
        self._logf = logf
        self._backend = backend
        TCPServer.__init__(self, addr, TCPSVNRequestHandler)



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Subversion wc library tests."""

import os

from subvertpy import (
    wc,
)
from subvertpy.tests import (
    SubversionTestCase,
    TestCase,
)


class VersionTest(TestCase):
    def test_version_length(self):
        self.assertEqual(4, len(wc.version()))

    def test_api_version_length(self):
        self.assertEqual(4, len(wc.api_version()))

    def test_api_version_later_same(self):
        self.assertTrue(wc.api_version() <= wc.version())


class AdmTests(TestCase):
    def test_get_adm_dir(self):
        self.assertEqual(".svn", wc.get_adm_dir())

    def test_set_adm_dir(self):
        old_dir_name = wc.get_adm_dir()
        try:
            wc.set_adm_dir(b"_svn")
            self.assertEqual("_svn", wc.get_adm_dir())
        finally:
            wc.set_adm_dir(old_dir_name)

    def test_is_normal_prop(self):
        self.assertTrue(wc.is_normal_prop("svn:ignore"))

    def test_is_normal_prop_false(self):
        self.assertFalse(wc.is_normal_prop("svn:entry:foo"))
        self.assertFalse(wc.is_normal_prop("svn:wc:foo"))

    def test_is_entry_prop(self):
        self.assertTrue(wc.is_entry_prop("svn:entry:foo"))

    def test_is_entry_prop_false(self):
        self.assertFalse(wc.is_entry_prop("svn:ignore"))

    def test_is_wc_prop(self):
        self.assertTrue(wc.is_wc_prop("svn:wc:foo"))

    def test_is_wc_prop_false(self):
        self.assertFalse(wc.is_wc_prop("svn:ignore"))

    def test_is_adm_dir(self):
        self.assertTrue(wc.is_adm_dir(".svn"))
        self.assertFalse(wc.is_adm_dir("foo"))

    def test_match_ignore_list(self):
        self.assertTrue(wc.match_ignore_list("foo", ["f*"]))
        self.assertTrue(wc.match_ignore_list("foo", ["foo"]))
        self.assertFalse(wc.match_ignore_list("foo", []))
        self.assertFalse(wc.match_ignore_list("foo", ["bar"]))

    def test_check_wc_nonexistent(self):
        import tempfile

        tmpdir = tempfile.mkdtemp()
        try:
            result = wc.check_wc(tmpdir)
            self.assertEqual(0, result)
        finally:
            import shutil

            shutil.rmtree(tmpdir)

    def test_get_actual_target(self):
        result = wc.get_actual_target("/foo/bar")
        self.assertIsInstance(result, tuple)
        self.assertEqual(2, len(result))


class WcTests(SubversionTestCase):
    def test_revision_status(self):
        self.make_client("repos", "checkout")
        ret = wc.revision_status("checkout")
        self.assertEqual((0, 0, 0, 0), ret)

    def test_revision_status_trailing(self):
        self.make_client("repos", "checkout")
        ret = wc.revision_status("checkout/")
        self.assertEqual((0, 0, 0, 0), ret)

    def test_revision_status_committed(self):
        self.make_client("repos", "checkout")
        ret = wc.revision_status("checkout", committed=True)
        self.assertEqual((0, 0, 0, 0), ret)

    def test_revision_status_trail_url(self):
        self.make_client("repos", "checkout")
        ret = wc.revision_status("checkout", trail_url=None)
        self.assertEqual((0, 0, 0, 0), ret)


class ContextTests(SubversionTestCase):
    def setUp(self):
        super(ContextTests, self).setUp()
        self.repos_url = self.make_client("repos", "checkout")

    def test_create_context(self):
        ctx = wc.Context()
        self.assertIsNotNone(ctx)

    def test_locked(self):
        ctx = wc.Context()
        result = ctx.locked(os.path.abspath("checkout"))
        self.assertIsInstance(result, tuple)
        self.assertEqual(2, len(result))

    def test_text_modified(self):
        ctx = wc.Context()
        self.build_tree({"checkout/foo": b"content"})
        self.client_add("checkout/foo")
        self.client_commit("checkout", message="add foo")
        # Unmodified file
        result = ctx.text_modified(os.path.abspath("checkout/foo"))
        self.assertFalse(result)
        # Modify the file
        self.build_tree({"checkout/foo": b"changed"})
        result = ctx.text_modified(os.path.abspath("checkout/foo"))
        self.assertTrue(result)

    def test_props_modified(self):
        ctx = wc.Context()
        self.build_tree({"checkout/bar": b"content"})
        self.client_add("checkout/bar")
        self.client_commit("checkout", message="add bar")
        result = ctx.props_modified(os.path.abspath("checkout/bar"))
        self.assertFalse(result)

    def test_conflicted(self):
        ctx = wc.Context()
        self.build_tree({"checkout/baz": b"content"})
        self.client_add("checkout/baz")
        self.client_commit("checkout", message="add baz")
        result = ctx.conflicted(os.path.abspath("checkout/baz"))
        # Returns a tuple of (text_conflicted, prop_conflicted, tree_conflicted)
        self.assertIsInstance(result, tuple)

    def test_status(self):
        ctx = wc.Context()
        self.build_tree({"checkout/statusfile": b"content"})
        self.client_add("checkout/statusfile")
        result = ctx.status(os.path.abspath("checkout/statusfile"))
        self.assertIsNotNone(result)

    def test_get_prop_diffs(self):
        ctx = wc.Context()
        self.build_tree({"checkout/propfile": b"content"})
        self.client_add("checkout/propfile")
        self.client_commit("checkout", message="add")
        result = ctx.get_prop_diffs(os.path.abspath("checkout/propfile"))
        self.assertIsInstance(result, tuple)
        self.assertEqual(2, len(result))

    def test_walk_status(self):
        ctx = wc.Context()
        self.build_tree({"checkout/walkfile": b"content"})
        self.client_add("checkout/walkfile")
        self.client_commit("checkout", message="add walkfile")
        statuses = []

        def receiver(path, status):
            statuses.append((path, status))

        ctx.walk_status(os.path.abspath("checkout"), receiver)
        self.assertTrue(len(statuses) > 0)

    def test_walk_status_with_options(self):
        from subvertpy import ra

        ctx = wc.Context()
        self.build_tree({"checkout/walkopt": b"content"})
        self.client_add("checkout/walkopt")
        self.client_commit("checkout", message="add walkopt")
        statuses = []

        def receiver(path, status):
            statuses.append((path, status))

        ctx.walk_status(
            os.path.abspath("checkout"),
            receiver,
            depth=ra.DEPTH_INFINITY,
            get_all=True,
            no_ignore=True,
        )
        self.assertTrue(len(statuses) > 0)

    def test_walk_status_ignore_patterns(self):
        ctx = wc.Context()
        self.build_tree({"checkout/walkign": b"content"})
        self.client_add("checkout/walkign")
        self.client_commit("checkout", message="add walkign")
        statuses = []

        def receiver(path, status):
            statuses.append((path, status))

        ctx.walk_status(
            os.path.abspath("checkout"), receiver, ignore_patterns=["*.pyc", "*.o"]
        )
        self.assertTrue(len(statuses) > 0)

    def test_walk_status_ignore_text_mode(self):
        ctx = wc.Context()
        self.build_tree({"checkout/walktxt": b"content"})
        self.client_add("checkout/walktxt")
        self.client_commit("checkout", message="add walktxt")
        statuses = []

        def receiver(path, status):
            statuses.append((path, status))

        ctx.walk_status(os.path.abspath("checkout"), receiver, ignore_text_mode=True)
        self.assertTrue(len(statuses) > 0)

    def test_crawl_revisions_with_options(self):
        ctx = wc.Context()
        self.build_tree({"checkout/crawlopt": b"data"})
        self.client_add("checkout/crawlopt")
        self.client_commit("checkout", message="add crawlopt")

        class MockReporter:
            def __init__(self):
                self.calls = []

            def set_path(self, path, revision, start_empty, lock_token, depth):
                self.calls.append(("set_path", path))

            def finish(self):
                self.calls.append(("finish",))

            def delete_path(self, path):
                pass

            def link_path(self, path, url, revision, start_empty, lock_token, depth):
                pass

            def abort(self):
                pass

        reporter = MockReporter()
        ctx.crawl_revisions(
            os.path.abspath("checkout"),
            reporter,
            restore_files=True,
            honor_depth_exclude=False,
            depth_compatibility_trick=True,
            use_commit_times=True,
        )
        call_types = [c[0] for c in reporter.calls]
        self.assertIn("set_path", call_types)
        self.assertIn("finish", call_types)

    def test_add_from_disk_no_lock(self):
        ctx = wc.Context()
        self.build_tree({"checkout/diskfile": b"disk content"})
        from subvertpy import SubversionException

        # add_from_disk requires a write lock
        self.assertRaises(
            SubversionException, ctx.add_from_disk, os.path.abspath("checkout/diskfile")
        )

    def test_add_lock_requires_write_lock(self):
        ctx = wc.Context()
        self.build_tree({"checkout/lockwc": b"data"})
        self.client_add("checkout/lockwc")
        self.client_commit("checkout", message="add lockwc")
        lock = wc.Lock(token="opaquelocktoken:test-token")
        from subvertpy import SubversionException

        # add_lock requires a write lock on the WC
        self.assertRaises(
            SubversionException, ctx.add_lock, os.path.abspath("checkout/lockwc"), lock
        )

    def test_remove_lock_requires_write_lock(self):
        ctx = wc.Context()
        self.build_tree({"checkout/rmlockwc": b"data"})
        self.client_add("checkout/rmlockwc")
        self.client_commit("checkout", message="add rmlockwc")
        from subvertpy import SubversionException

        self.assertRaises(
            SubversionException, ctx.remove_lock, os.path.abspath("checkout/rmlockwc")
        )

    def test_crawl_revisions(self):
        ctx = wc.Context()
        self.build_tree({"checkout/crawlfile": b"data"})
        self.client_add("checkout/crawlfile")
        self.client_commit("checkout", message="add crawlfile")

        class MockReporter:
            def __init__(self):
                self.calls = []

            def set_path(self, path, revision, start_empty, lock_token, depth):
                self.calls.append(("set_path", path, revision, start_empty))

            def finish(self):
                self.calls.append(("finish",))

            def delete_path(self, path):
                self.calls.append(("delete_path", path))

            def link_path(self, path, url, revision, start_empty, lock_token, depth):
                self.calls.append(("link_path", path, url))

            def abort(self):
                self.calls.append(("abort",))

        reporter = MockReporter()
        ctx.crawl_revisions(os.path.abspath("checkout"), reporter, restore_files=False)
        self.assertTrue(len(reporter.calls) > 0)
        # Should have called set_path at least once and finish
        call_types = [c[0] for c in reporter.calls]
        self.assertIn("set_path", call_types)
        self.assertIn("finish", call_types)

    def test_get_update_editor(self):
        ctx = wc.Context()
        self.build_tree({"checkout/updfile": b"data"})
        self.client_add("checkout/updfile")
        self.client_commit("checkout", message="add updfile")
        editor = ctx.get_update_editor(os.path.abspath("checkout"), "")
        self.assertIsNotNone(editor)
        editor.abort()

    def test_get_update_editor_with_options(self):
        from subvertpy import ra

        ctx = wc.Context()
        self.build_tree({"checkout/upd2": b"data"})
        self.client_add("checkout/upd2")
        self.client_commit("checkout", message="add upd2")
        editor = ctx.get_update_editor(
            os.path.abspath("checkout"),
            "",
            use_commit_times=True,
            depth=ra.DEPTH_INFINITY,
            depth_is_sticky=False,
            allow_unver_obstructions=False,
            adds_as_modification=True,
            server_performs_filtering=False,
            clean_checkout=False,
        )
        self.assertIsNotNone(editor)
        editor.abort()

    def test_get_update_editor_preserved_exts(self):
        ctx = wc.Context()
        self.build_tree({"checkout/upd3": b"data"})
        self.client_add("checkout/upd3")
        self.client_commit("checkout", message="add upd3")
        editor = ctx.get_update_editor(
            os.path.abspath("checkout"), "", preserved_exts=[".mine", ".theirs"]
        )
        self.assertIsNotNone(editor)
        editor.abort()

    def test_get_update_editor_notify_func(self):
        ctx = wc.Context()
        self.build_tree({"checkout/upd4": b"data"})
        self.client_add("checkout/upd4")
        self.client_commit("checkout", message="add upd4")
        notifications = []

        def notify(info):
            notifications.append(info)

        editor = ctx.get_update_editor(
            os.path.abspath("checkout"), "", notify_func=notify
        )
        self.assertIsNotNone(editor)
        editor.abort()

    def test_get_update_editor_conflict_func_not_implemented(self):
        ctx = wc.Context()
        self.build_tree({"checkout/upd5": b"data"})
        self.client_add("checkout/upd5")
        self.client_commit("checkout", message="add upd5")
        self.assertRaises(
            NotImplementedError,
            ctx.get_update_editor,
            os.path.abspath("checkout"),
            "",
            conflict_func=lambda: None,
        )

    def test_get_update_editor_external_func_not_implemented(self):
        ctx = wc.Context()
        self.build_tree({"checkout/upd6": b"data"})
        self.client_add("checkout/upd6")
        self.client_commit("checkout", message="add upd6")
        self.assertRaises(
            NotImplementedError,
            ctx.get_update_editor,
            os.path.abspath("checkout"),
            "",
            external_func=lambda: None,
        )

    def test_get_update_editor_dirents_func_not_implemented(self):
        ctx = wc.Context()
        self.build_tree({"checkout/upd7": b"data"})
        self.client_add("checkout/upd7")
        self.client_commit("checkout", message="add upd7")
        self.assertRaises(
            NotImplementedError,
            ctx.get_update_editor,
            os.path.abspath("checkout"),
            "",
            dirents_func=lambda: None,
        )

    def test_add_from_disk_with_props(self):
        ctx = wc.Context()
        self.build_tree({"checkout/diskprops": b"data"})
        from subvertpy import SubversionException

        # add_from_disk requires a write lock, but we can test that
        # the props parameter is accepted
        self.assertRaises(
            SubversionException,
            ctx.add_from_disk,
            os.path.abspath("checkout/diskprops"),
            props={b"svn:eol-style": b"native"},
        )

    def test_add_from_disk_skip_checks(self):
        ctx = wc.Context()
        self.build_tree({"checkout/diskskip": b"data"})
        from subvertpy import SubversionException

        # add_from_disk requires a write lock
        self.assertRaises(
            SubversionException,
            ctx.add_from_disk,
            os.path.abspath("checkout/diskskip"),
            skip_checks=True,
        )

    def test_add_from_disk_notify(self):
        ctx = wc.Context()
        self.build_tree({"checkout/disknot": b"data"})
        from subvertpy import SubversionException

        notifications = []

        def notify_cb(info):
            notifications.append(info)

        # add_from_disk requires a write lock
        self.assertRaises(
            SubversionException,
            ctx.add_from_disk,
            os.path.abspath("checkout/disknot"),
            notify=notify_cb,
        )

    def test_crawl_revisions_with_notify(self):
        ctx = wc.Context()
        self.build_tree({"checkout/crawlnot": b"data"})
        self.client_add("checkout/crawlnot")
        self.client_commit("checkout", message="add crawlnot")

        class MockReporter:
            def __init__(self):
                self.calls = []

            def set_path(self, path, revision, start_empty, lock_token, depth):
                self.calls.append(("set_path", path))

            def finish(self):
                self.calls.append(("finish",))

            def delete_path(self, path):
                pass

            def link_path(self, path, url, revision, start_empty, lock_token, depth):
                pass

            def abort(self):
                pass

        reporter = MockReporter()
        notifications = []

        def notify_cb(info):
            notifications.append(info)

        ctx.crawl_revisions(
            os.path.abspath("checkout"), reporter, restore_files=True, notify=notify_cb
        )
        call_types = [c[0] for c in reporter.calls]
        self.assertIn("set_path", call_types)
        self.assertIn("finish", call_types)

    def test_ensure_adm(self):
        from subvertpy import repos as svn_repos

        repo = svn_repos.Repository("repos")
        uuid = repo.fs().get_uuid()
        ctx = wc.Context()
        # ensure_adm on an existing checkout should succeed
        ctx.ensure_adm(
            os.path.abspath("checkout"), self.repos_url, self.repos_url, uuid, 0
        )

    def test_ensure_adm_with_depth(self):
        from subvertpy import repos as svn_repos, ra

        repo = svn_repos.Repository("repos")
        uuid = repo.fs().get_uuid()
        ctx = wc.Context()
        ctx.ensure_adm(
            os.path.abspath("checkout"),
            self.repos_url,
            self.repos_url,
            uuid,
            0,
            depth=ra.DEPTH_INFINITY,
        )

    def test_committed_queue_create(self):
        queue = wc.CommittedQueue()
        self.assertIsNotNone(queue)

    def test_committed_queue_queue(self):
        ctx = wc.Context()
        self.build_tree({"checkout/qfile": b"queue data"})
        self.client_add("checkout/qfile")
        self.client_commit("checkout", message="add qfile")
        queue = wc.CommittedQueue()
        queue.queue(os.path.abspath("checkout/qfile"), ctx)

    def test_process_committed_queue_requires_write_lock(self):
        from subvertpy import SubversionException

        ctx = wc.Context()
        self.build_tree({"checkout/pcqfile": b"data"})
        self.client_add("checkout/pcqfile")
        self.client_commit("checkout", message="add pcqfile")
        queue = wc.CommittedQueue()
        queue.queue(os.path.abspath("checkout/pcqfile"), ctx)
        self.assertRaises(
            SubversionException,
            ctx.process_committed_queue,
            queue,
            1,
            "2026-01-01T00:00:00.000000Z",
            "testuser",
        )


class LockTests(TestCase):
    def test_create_lock(self):
        lock = wc.Lock()
        self.assertIsNotNone(lock)

    def test_create_lock_with_token(self):
        lock = wc.Lock(token="opaquelocktoken:test")
        self.assertEqual(b"opaquelocktoken:test", lock.token)

    def test_path_default(self):
        lock = wc.Lock()
        self.assertIsNone(lock.path)

    def test_set_path(self):
        lock = wc.Lock()
        lock.path = b"/some/path"
        self.assertEqual("/some/path", lock.path)

    def test_token_default(self):
        lock = wc.Lock()
        self.assertIsNone(lock.token)

    def test_set_token(self):
        lock = wc.Lock()
        lock.token = b"opaquelocktoken:abc"
        self.assertEqual(b"opaquelocktoken:abc", lock.token)


class PristineTests(SubversionTestCase):
    def setUp(self):
        super(PristineTests, self).setUp()
        self.repos_url = self.make_client("repos", "checkout")

    def test_get_pristine_contents(self):
        self.build_tree({"checkout/pristfile": b"pristine data"})
        self.client_add("checkout/pristfile")
        self.client_commit("checkout", message="add pristfile")
        stream = wc.get_pristine_contents(os.path.abspath("checkout/pristfile"))
        self.assertIsNotNone(stream)

    def test_get_pristine_copy_path(self):
        import warnings

        self.build_tree({"checkout/cpfile": b"data"})
        self.client_add("checkout/cpfile")
        self.client_commit("checkout", message="add cpfile")
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            result = wc.get_pristine_copy_path(os.path.abspath("checkout/cpfile"))
        self.assertIsNotNone(result)


class EnsureAdmTests(SubversionTestCase):
    def test_ensure_adm_bogus_url(self):
        from subvertpy import SubversionException

        self.make_client("repos", "checkout")
        self.assertRaises(
            SubversionException, wc.ensure_adm, "checkout", "fake-uuid", "file:///fake"
        )

    def test_ensure_adm_with_repos_and_rev(self):
        from subvertpy import repos as svn_repos

        repos_url = self.make_client("repos", "checkout")
        repo = svn_repos.Repository("repos")
        uuid = repo.fs().get_uuid()
        wc.ensure_adm("checkout", uuid, repos_url, repos=repos_url, rev=0)

    def test_ensure_adm_with_depth(self):
        from subvertpy import repos as svn_repos, ra

        repos_url = self.make_client("repos", "checkout")
        repo = svn_repos.Repository("repos")
        uuid = repo.fs().get_uuid()
        wc.ensure_adm(
            "checkout", uuid, repos_url, repos=repos_url, rev=0, depth=ra.DEPTH_INFINITY
        )


class CleanupTests(SubversionTestCase):
    def test_cleanup(self):
        self.make_client("repos", "checkout")
        # cleanup on a clean working copy should succeed
        wc.cleanup("checkout")

    def test_cleanup_with_diff3_cmd(self):
        self.make_client("repos", "checkout")
        wc.cleanup("checkout", diff3_cmd=None)


class ConstantsTests(TestCase):
    def test_schedule_constants(self):
        self.assertEqual(0, wc.SCHEDULE_NORMAL)
        self.assertEqual(1, wc.SCHEDULE_ADD)
        self.assertEqual(2, wc.SCHEDULE_DELETE)
        self.assertEqual(3, wc.SCHEDULE_REPLACE)

    def test_conflict_choose_constants(self):
        self.assertIsInstance(wc.CONFLICT_CHOOSE_POSTPONE, int)
        self.assertIsInstance(wc.CONFLICT_CHOOSE_BASE, int)
        self.assertIsInstance(wc.CONFLICT_CHOOSE_THEIRS_FULL, int)
        self.assertIsInstance(wc.CONFLICT_CHOOSE_MINE_FULL, int)
        self.assertIsInstance(wc.CONFLICT_CHOOSE_THEIRS_CONFLICT, int)
        self.assertIsInstance(wc.CONFLICT_CHOOSE_MINE_CONFLICT, int)
        self.assertIsInstance(wc.CONFLICT_CHOOSE_MERGED, int)

    def test_status_constants(self):
        self.assertIsInstance(wc.STATUS_NONE, int)
        self.assertIsInstance(wc.STATUS_UNVERSIONED, int)
        self.assertIsInstance(wc.STATUS_NORMAL, int)
        self.assertIsInstance(wc.STATUS_ADDED, int)
        self.assertIsInstance(wc.STATUS_MISSING, int)
        self.assertIsInstance(wc.STATUS_DELETED, int)
        self.assertIsInstance(wc.STATUS_REPLACED, int)
        self.assertIsInstance(wc.STATUS_MODIFIED, int)
        self.assertIsInstance(wc.STATUS_MERGED, int)
        self.assertIsInstance(wc.STATUS_CONFLICTED, int)
        self.assertIsInstance(wc.STATUS_IGNORED, int)
        self.assertIsInstance(wc.STATUS_OBSTRUCTED, int)
        self.assertIsInstance(wc.STATUS_EXTERNAL, int)
        self.assertIsInstance(wc.STATUS_INCOMPLETE, int)

    def test_translate_constants(self):
        self.assertIsInstance(wc.TRANSLATE_FROM_NF, int)
        self.assertIsInstance(wc.TRANSLATE_TO_NF, int)
        self.assertIsInstance(wc.TRANSLATE_FORCE_EOL_REPAIR, int)
        self.assertIsInstance(wc.TRANSLATE_NO_OUTPUT_CLEANUP, int)
        self.assertIsInstance(wc.TRANSLATE_FORCE_COPY, int)
        self.assertIsInstance(wc.TRANSLATE_USE_GLOBAL_TMP, int)



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Subversion client library tests."""

from datetime import datetime, timedelta
import os
from io import BytesIO
import shutil
import tempfile

from subvertpy import (
    SubversionException,
    client,
    ra,
)
from subvertpy.tests import (
    SubversionTestCase,
    TestCase,
)


class VersionTest(TestCase):
    def test_version_length(self):
        self.assertEqual(4, len(client.version()))

    def test_api_version_length(self):
        self.assertEqual(4, len(client.api_version()))

    def test_api_version_later_same(self):
        self.assertTrue(client.api_version() <= client.version())


class TestClient(SubversionTestCase):
    def setUp(self):

        super(TestClient, self).setUp()
        self.repos_url = self.make_client("d", "dc")
        self.client = client.Client(auth=ra.Auth([ra.get_username_provider()]))

    def tearDown(self):
        del self.client
        super(TestClient, self).tearDown()

    def test_add(self):
        self.build_tree({"dc/foo": None})
        self.client.add("dc/foo")

    def test_commit(self):
        self.build_tree({"dc/foo": None})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Amessage"
        self.client.commit(["dc"])
        r = ra.RemoteAccess(self.repos_url)
        revprops = r.rev_proplist(1)
        self.assertEqual(b"Amessage", revprops["svn:log"])

    def test_commit_start(self):
        self.build_tree({"dc/foo": None})
        self.client = client.Client(
            auth=ra.Auth([ra.get_username_provider()]),
            log_msg_func=lambda c: "Bmessage",
        )
        self.client.add("dc/foo")
        self.client.commit(["dc"])
        r = ra.RemoteAccess(self.repos_url)
        revprops = r.rev_proplist(1)
        self.assertEqual(b"Bmessage", revprops["svn:log"])

    def test_mkdir(self):
        self.client.mkdir(["dc/foo"])
        self.client.mkdir("dc/bar")
        self.client.mkdir("dc/bla", revprops={"svn:log": "foo"})

    def test_export(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.commit(["dc"])
        self.client.export(self.repos_url, "de")
        self.assertEqual(["foo"], os.listdir("de"))

    def test_export_new_option(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.commit(["dc"])
        self.client.export(
            self.repos_url, "de", ignore_externals=True, ignore_keywords=True
        )
        self.assertEqual(["foo"], os.listdir("de"))

    def test_set_config(self):
        config = client.get_config()
        self.client.config = config

    def test_set_config_none(self):
        self.client.config = None

    def test_get_config(self):
        self.assertIsInstance(client.get_config(), client.Config)
        try:
            base_dir = tempfile.mkdtemp()
            base_dir_basename = os.path.basename(base_dir)
            svn_cfg_dir = os.path.join(base_dir, ".subversion")
            os.mkdir(svn_cfg_dir)
            with open(os.path.join(svn_cfg_dir, "config"), "w") as svn_cfg:
                svn_cfg.write("[miscellany]\n")
                svn_cfg.write("global-ignores = %s" % base_dir_basename)
            config = client.get_config(svn_cfg_dir)
            self.assertIsInstance(config, client.Config)
            ignores = config.get_default_ignores()
            self.assertTrue(
                base_dir_basename.encode("utf-8") in ignores,
                "no %r in %r" % (base_dir_basename, ignores),
            )
        finally:
            shutil.rmtree(base_dir)

    def test_diff(self):
        dc = self.get_commit_editor(self.repos_url)
        f = dc.add_file("foo")
        f.modify(b"foo1")
        dc.close()

        dc = self.get_commit_editor(self.repos_url)
        f = dc.open_file("foo")
        f.modify(b"foo2")
        dc.close()

        (outf, errf) = self.client.diff(1, 2, self.repos_url, self.repos_url)
        self.addCleanup(outf.close)
        self.addCleanup(errf.close)
        self.assertEqual(
            b"""Index: foo
===================================================================
--- foo\t(revision 1)
+++ foo\t(revision 2)
@@ -1 +1 @@
-foo1
\\ No newline at end of file
+foo2
\\ No newline at end of file
""".splitlines(),
            outf.read().splitlines(),
        )
        self.assertEqual(b"", errf.read())

    def assertCatEquals(self, value, revision=None):
        io = BytesIO()
        self.client.cat("dc/foo", io, revision)
        self.assertEqual(value, io.getvalue())

    def test_cat(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.assertCatEquals(b"bla")
        self.build_tree({"dc/foo": b"blabla"})
        self.client.commit(["dc"])
        self.assertCatEquals(b"blabla")
        self.assertCatEquals(b"bla", revision=1)
        self.assertCatEquals(b"blabla", revision=2)

    def assertLogEntryChangedPathsEquals(self, expected, entry):
        changed_paths = entry["changed_paths"]
        self.assertIsInstance(changed_paths, dict)
        self.assertEqual(sorted(expected), sorted(changed_paths.keys()))

    def assertLogEntryMessageEquals(self, expected, entry):
        self.assertEqual(expected, entry["revprops"]["svn:log"])

    def assertLogEntryDateAlmostEquals(self, expected, entry, delta):
        actual = datetime.strptime(
            entry["revprops"]["svn:date"].decode("utf-8"), "%Y-%m-%dT%H:%M:%S.%fZ"
        )
        self.assertTrue((actual - expected) < delta)

    def test_log(self):
        log_entries = []
        commit_msg_1 = b"Commit"
        commit_msg_2 = b"Commit 2"
        delta = timedelta(hours=1)

        def cb(changed_paths, revision, revprops, has_children=False):
            log_entries.append(
                {
                    "changed_paths": changed_paths,
                    "revision": revision,
                    "revprops": revprops,
                    "has_children": has_children,
                }
            )

        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: commit_msg_1
        self.client.commit(["dc"])
        commit_1_dt = datetime.utcnow()
        self.client.log(cb, "dc/foo", start_rev="HEAD", end_rev=1)
        self.assertEqual(1, len(log_entries))
        self.assertEqual(None, log_entries[0]["changed_paths"])
        self.assertEqual(1, log_entries[0]["revision"])
        self.assertLogEntryMessageEquals(commit_msg_1, log_entries[0])
        self.assertLogEntryDateAlmostEquals(commit_1_dt, log_entries[0], delta)
        self.build_tree(
            {
                "dc/foo": b"blabla",
                "dc/bar": b"blablabla",
            }
        )
        self.client.add("dc/bar")
        self.client.log_msg_func = lambda c: commit_msg_2
        self.client.commit(["dc"])
        commit_2_dt = datetime.utcnow()
        log_entries = []
        self.client.log(
            cb, "dc/foo", start_rev="HEAD", end_rev=1, discover_changed_paths=True
        )
        self.assertEqual(2, len(log_entries))
        self.assertLogEntryChangedPathsEquals(["/foo", "/bar"], log_entries[0])
        self.assertEqual(2, log_entries[0]["revision"])
        self.assertLogEntryMessageEquals(commit_msg_2, log_entries[0])
        self.assertLogEntryDateAlmostEquals(commit_2_dt, log_entries[0], delta)
        self.assertLogEntryChangedPathsEquals(["/foo"], log_entries[1])
        self.assertEqual(1, log_entries[1]["revision"])
        self.assertLogEntryMessageEquals(commit_msg_1, log_entries[1])
        self.assertLogEntryDateAlmostEquals(commit_1_dt, log_entries[1], delta)
        log_entries = []
        self.client.log(
            cb, "dc/foo", start_rev=2, end_rev=2, discover_changed_paths=True
        )
        self.assertEqual(1, len(log_entries))
        self.assertLogEntryChangedPathsEquals(["/foo", "/bar"], log_entries[0])
        self.assertEqual(2, log_entries[0]["revision"])
        self.assertLogEntryMessageEquals(commit_msg_2, log_entries[0])
        self.assertLogEntryDateAlmostEquals(commit_2_dt, log_entries[0], delta)

    def test_info(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        info = self.client.info("dc/foo")
        self.assertEqual(["foo"], list(info.keys()))
        self.assertEqual(1, info["foo"].revision)
        self.assertEqual(3, info["foo"].size)
        self.build_tree({"dc/bar": b"blablabla"})
        self.client.add(os.path.abspath("dc/bar"))

    def test_info_nonexistant(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.assertRaises(SubversionException, self.client.info, "dc/missing")

    def test_set_get_prop_with_path(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client_add("dc/foo")
        self.client_set_prop("dc/foo", "svn:eol-style", "native")
        self.client_commit("dc", message="Commit")
        self.assertEqual(
            self.client_get_prop("dc/foo", "svn:eol-style", "HEAD"), b"native"
        )

    def test_set_get_prop_with_url(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client_add("dc/foo")
        self.client_set_prop("dc/foo", "svn:eol-style", "native")
        self.client_commit("dc", message="Commit")
        self.assertEqual(
            self.client_get_prop(self.repos_url + "/foo", "svn:eol-style", "HEAD"),
            b"native",
        )

    def test_checkout(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        checkout_dir = os.path.join(self.test_dir, "checkout2")
        self.client.checkout(self.repos_url, checkout_dir, "HEAD")
        self.assertTrue(os.path.exists(os.path.join(checkout_dir, "foo")))

    def test_delete(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.delete(["dc/foo"])
        self.assertFalse(os.path.exists("dc/foo"))

    def test_copy(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.copy("dc/foo", "dc/bar")
        self.assertTrue(os.path.exists("dc/bar"))

    def test_propset_propget(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.propset("myprop", "myval", "dc/foo", False, True)
        ret = self.client.propget("myprop", "dc/foo", "WORKING", "WORKING")
        self.assertIsInstance(ret, dict)
        self.assertIn(b"myval", ret.values())

    def test_propget_url(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.propset("myprop", "myval", "dc/foo", False, True)
        self.client.commit(["dc"])
        url = self.repos_url + "/foo"
        ret = self.client.propget("myprop", url, 2, 2)
        self.assertIsInstance(ret, dict)
        self.assertIn(b"myval", ret.values())

    def test_propget_abspath(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.propset("myprop", "myval", "dc/foo", False, True)
        abspath = os.path.abspath("dc/foo")
        ret = self.client.propget("myprop", abspath, "WORKING", "WORKING")
        self.assertIsInstance(ret, dict)
        self.assertIn(b"myval", ret.values())

    def test_proplist(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.propset("myprop", "myval", "dc/foo", False, True)
        self.client.commit(["dc"])
        result = self.client.proplist("dc/foo", "WORKING", 0)
        self.assertIsInstance(result, list)

    def test_update(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.update(["dc"], "HEAD")

    def test_list(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        entries = self.client.list(self.repos_url, "HEAD", 0)
        self.assertIsInstance(entries, dict)

    def test_config(self):
        config = client.get_config()
        self.assertIsInstance(config, client.Config)

    def test_config_default_ignores(self):
        config = client.get_config()
        ignores = config.get_default_ignores()
        self.assertIsInstance(ignores, list)

    def test_lock_unlock(self):
        self.build_tree({"dc/foo": b"bla"})
        self.client.add("dc/foo")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.lock(["dc/foo"], "test lock comment")
        self.client.unlock(["dc/foo"])

    def test_resolve(self):
        self.build_tree({"dc/resolveme": b"content"})
        self.client.add("dc/resolveme")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        # resolve on a non-conflicted path should succeed without error
        self.client.resolve("dc/resolveme", 0, 0)

    def test_add_with_options(self):
        self.build_tree({"dc/addopts": b"data"})
        self.client.add(
            "dc/addopts",
            recursive=True,
            force=False,
            no_ignore=True,
            add_parents=False,
            no_autoprops=True,
        )

    def test_commit_with_options(self):
        self.build_tree({"dc/commitopt": b"data"})
        self.client.add("dc/commitopt")
        self.client.log_msg_func = lambda c: "Commit with opts"
        callbacks = []
        self.client.commit(
            ["dc"],
            recurse=True,
            keep_locks=False,
            keep_changelist=False,
            commit_as_operations=True,
            include_file_externals=False,
            include_dir_externals=False,
            callback=lambda *args: callbacks.append(args),
        )
        self.assertEqual(1, len(callbacks))

    def test_commit_revprops(self):
        self.build_tree({"dc/rptest": b"data"})
        self.client.add("dc/rptest")
        self.client.log_msg_func = lambda c: "revprop commit"
        self.client.commit(["dc"], revprops={"custom:testprop": "testval"})

    def test_update_with_options(self):
        self.build_tree({"dc/updopt": b"data"})
        self.client.add("dc/updopt")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.update(
            ["dc"],
            "HEAD",
            recurse=True,
            ignore_externals=True,
            depth_is_sticky=False,
            allow_unver_obstructions=False,
            adds_as_modification=True,
            make_parents=False,
        )

    def test_checkout_with_options(self):
        self.build_tree({"dc/chkopt": b"data"})
        self.client.add("dc/chkopt")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        checkout_dir = os.path.join(self.test_dir, "checkout3")
        self.client.checkout(
            self.repos_url,
            checkout_dir,
            "HEAD",
            peg_rev="HEAD",
            recurse=True,
            allow_unver_obstructions=False,
        )
        self.assertTrue(os.path.exists(os.path.join(checkout_dir, "chkopt")))

    def test_export_with_options(self):
        self.build_tree({"dc/expopt": b"data"})
        self.client.add("dc/expopt")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        export_dir = os.path.join(self.test_dir, "export_opts")
        self.client.export(
            self.repos_url,
            export_dir,
            rev="HEAD",
            peg_rev="HEAD",
            recurse=True,
            overwrite=False,
        )
        self.assertTrue(os.path.exists(os.path.join(export_dir, "expopt")))

    def test_diff_with_options(self):
        dc = self.get_commit_editor(self.repos_url)
        f = dc.add_file("diffopt")
        f.modify(b"v1")
        dc.close()
        dc = self.get_commit_editor(self.repos_url)
        f = dc.open_file("diffopt")
        f.modify(b"v2")
        dc.close()
        (outf, errf) = self.client.diff(
            1,
            2,
            self.repos_url,
            self.repos_url,
            ignore_ancestry=True,
            no_diff_deleted=False,
            ignore_content_type=False,
        )
        self.addCleanup(outf.close)
        self.addCleanup(errf.close)
        out = outf.read()
        self.assertIn(b"diffopt", out)

    def test_log_with_options(self):
        entries = []

        def cb(changed_paths, revision, revprops, has_children=False):
            entries.append(revision)

        self.build_tree({"dc/logopt": b"data"})
        self.client.add("dc/logopt")
        self.client.log_msg_func = lambda c: "Commit 1"
        self.client.commit(["dc"])
        self.build_tree({"dc/logopt": b"data2"})
        self.client.commit(["dc"])
        self.client.log(
            cb,
            "dc/logopt",
            start_rev="HEAD",
            end_rev=1,
            limit=1,
            discover_changed_paths=True,
            strict_node_history=True,
            include_merged_revisions=False,
        )
        self.assertEqual(1, len(entries))

    def test_cat_with_peg_revision(self):
        self.build_tree({"dc/pegcat": b"original"})
        self.client.add("dc/pegcat")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        io = BytesIO()
        self.client.cat("dc/pegcat", io, revision=1, peg_revision=1)
        self.assertEqual(b"original", io.getvalue())

    def test_delete_keep_local(self):
        self.build_tree({"dc/keepme": b"data"})
        self.client.add("dc/keepme")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.delete(["dc/keepme"], keep_local=True)
        # File should still exist on disk
        self.assertTrue(os.path.exists("dc/keepme"))

    def test_copy_with_options(self):
        self.build_tree({"dc/cpsrc": b"data"})
        self.client.add("dc/cpsrc")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.copy(
            "dc/cpsrc",
            "dc/cpdst",
            copy_as_child=False,
            make_parents=False,
            metadata_only=False,
        )
        self.assertTrue(os.path.exists("dc/cpdst"))

    def test_info_with_options(self):
        self.build_tree({"dc/infoopt": b"data"})
        self.client.add("dc/infoopt")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        info = self.client.info(
            "dc/infoopt",
            revision=1,
            peg_revision=1,
            depth=0,
            fetch_excluded=False,
            fetch_actual_only=False,
        )
        self.assertIn("infoopt", info)

    def test_lock_steal(self):
        self.build_tree({"dc/locksteal": b"data"})
        self.client.add("dc/locksteal")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.lock(["dc/locksteal"], "lock comment", steal_lock=True)
        self.client.unlock(["dc/locksteal"], break_lock=True)

    def test_mkdir_make_parents(self):
        self.client.mkdir("dc/parent/child", make_parents=True)
        self.assertTrue(os.path.exists("dc/parent/child"))

    def test_mkdir_with_callback(self):
        commits = []
        self.client.mkdir(
            [self.repos_url + "/remotedir"], callback=lambda *args: commits.append(args)
        )

    def test_propset_skip_checks(self):
        self.build_tree({"dc/propskip": b"data"})
        self.client.add("dc/propskip")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.propset("custom:myprop", b"value", "dc/propskip", skip_checks=True)

    def test_propset_with_options(self):
        self.build_tree({"dc/propopt": b"data"})
        self.client.add("dc/propopt")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.propset("myprop", "myval", "dc/propopt", False, True)
        ret = self.client.propget("myprop", "dc/propopt", "WORKING", "WORKING", True)
        self.assertIsInstance(ret, dict)

    def test_cat_expand_keywords(self):
        self.build_tree({"dc/kwcat": b"$Id$"})
        self.client.add("dc/kwcat")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        io = BytesIO()
        self.client.cat("dc/kwcat", io, revision=1, expand_keywords=False)
        self.assertEqual(b"$Id$", io.getvalue())

    def test_export_native_eol(self):
        self.build_tree({"dc/eolfile": b"line\n"})
        self.client.add("dc/eolfile")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        export_dir = os.path.join(self.test_dir, "export_eol")
        self.client.export(self.repos_url, export_dir, native_eol="LF")
        self.assertTrue(os.path.exists(os.path.join(export_dir, "eolfile")))

    def test_list_include_externals(self):
        self.build_tree({"dc/listfile": b"data"})
        self.client.add("dc/listfile")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        entries = self.client.list(self.repos_url, "HEAD", 0, include_externals=False)
        self.assertIsInstance(entries, dict)

    def test_delete_with_callback(self):
        self.build_tree({"dc/delcb": b"data"})
        self.client.add("dc/delcb")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        commits = []
        self.client.delete(
            [self.repos_url + "/delcb"],
            force=True,
            callback=lambda *args: commits.append(args),
        )

    def test_copy_with_src_rev(self):
        self.build_tree({"dc/cprev": b"data"})
        self.client.add("dc/cprev")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.copy("dc/cprev", "dc/cprev2", src_rev=1)
        self.assertTrue(os.path.exists("dc/cprev2"))

    def test_get_config_with_dir(self):
        import tempfile

        cfg_dir = tempfile.mkdtemp()
        try:
            config = client.get_config(cfg_dir)
            self.assertIsInstance(config, client.Config)
        finally:
            import shutil

            shutil.rmtree(cfg_dir)

    def test_copy_pin_externals(self):
        self.build_tree({"dc/cppin": b"data"})
        self.client.add("dc/cppin")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.copy("dc/cppin", "dc/cppin2", pin_externals=False)
        self.assertTrue(os.path.exists("dc/cppin2"))

    def test_copy_metadata_only(self):
        self.build_tree({"dc/cpmeta": b"data"})
        self.client.add("dc/cpmeta")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.copy("dc/cpmeta", "dc/cpmeta2", metadata_only=False)
        self.assertTrue(os.path.exists("dc/cpmeta2"))

    def test_copy_with_callback(self):
        self.build_tree({"dc/cpcb": b"data"})
        self.client.add("dc/cpcb")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.client.copy(
            self.repos_url + "/cpcb",
            self.repos_url + "/cpcb2",
            callback=lambda *args: None,
        )

    def test_diff_relative_to_dir(self):
        self.build_tree({"dc/diffrel": b"line1\n"})
        self.client.add("dc/diffrel")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.build_tree({"dc/diffrel": b"line1\nline2\n"})
        (outfile, errfile) = self.client.diff(
            1, "WORKING", "dc", "dc", relative_to_dir="dc"
        )
        out = outfile.read()
        self.assertIn(b"line2", out)

    def test_diff_encoding(self):
        self.build_tree({"dc/diffenc": b"hello\n"})
        self.client.add("dc/diffenc")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.build_tree({"dc/diffenc": b"hello\nworld\n"})
        (outfile, errfile) = self.client.diff(
            1, "WORKING", "dc", "dc", encoding="utf-8"
        )
        out = outfile.read()
        self.assertIn(b"world", out)

    def test_diff_ignore_content_type(self):
        self.build_tree({"dc/diffict": b"hello\n"})
        self.client.add("dc/diffict")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.build_tree({"dc/diffict": b"hello\nworld\n"})
        (outfile, errfile) = self.client.diff(
            1, "WORKING", "dc", "dc", ignore_content_type=True
        )
        out = outfile.read()
        self.assertIn(b"world", out)

    def test_diff_diffopts(self):
        self.build_tree({"dc/diffopt": b"hello\n"})
        self.client.add("dc/diffopt")
        self.client.log_msg_func = lambda c: "Commit"
        self.client.commit(["dc"])
        self.build_tree({"dc/diffopt": b"hello\nworld\n"})
        (outfile, errfile) = self.client.diff(1, "WORKING", "dc", "dc", diffopts=["-u"])
        out = outfile.read()
        self.assertIn(b"world", out)

    def test_notify_func_set_get(self):
        def notify_cb(info):
            pass

        self.assertIsNone(self.client.notify_func)
        self.client.notify_func = notify_cb
        self.assertIs(notify_cb, self.client.notify_func)
        self.client.notify_func = None
        self.assertIsNone(self.client.notify_func)



#!/usr/bin/python
# Demonstrates how to use the replay function to fetch the
# changes made in a revision.

from subvertpy.ra import RemoteAccess, Auth, get_username_provider

conn = RemoteAccess(
    "svn://svn.gnome.org/svn/gnome-specimen/trunk", auth=Auth([get_username_provider()])
)


class MyFileEditor:
    def change_prop(self, key, value):
        print("Change prop: %s -> %r" % (key, value))

    def apply_textdelta(self, base_checksum):
        # This should return a function that can receive delta windows
        def apply_window(x):
            pass

        return apply_window

    def close(self):
        pass


class MyDirEditor:
    def open_directory(self, *args):
        print("Open dir: %s (base revnum: %r)" % args)
        return MyDirEditor()

    def add_directory(self, path, copyfrom_path=None, copyfrom_rev=-1):
        print("Add dir: %s (from %r:%r)" % (path, copyfrom_path, copyfrom_rev))
        return MyDirEditor()

    def open_file(self, *args):
        print("Open file: %s (base revnum: %r)" % args)
        return MyFileEditor()

    def add_file(self, path, copyfrom_path=None, copyfrom_rev=-1):
        print("Add file: %s (from %r:%r)" % (path, copyfrom_path, copyfrom_rev))
        return MyFileEditor()

    def change_prop(self, key, value):
        print("Change prop %s -> %r" % (key, value))

    def delete_entry(self, path, revision):
        print("Delete: %s" % path)

    def close(self):
        pass


class MyEditor:
    def set_target_revision(self, revnum):
        print("Target revision: %d" % revnum)

    def abort(self):
        print("Aborted")

    def close(self):
        print("Closed")

    def open_root(self, base_revnum):
        print("/")
        return MyDirEditor()


editor = MyEditor()
conn.replay(230, 1, editor)



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for subvertpy.delta."""

from io import BytesIO

from subvertpy.delta import (
    apply_txdelta_handler,
    apply_txdelta_handler_chunks,
    apply_txdelta_window,
    decode_length,
    encode_length,
    pack_svndiff0,
    pack_svndiff0_window,
    pack_svndiff_instruction,
    send_stream,
    txdelta_apply_ops,
    unpack_svndiff0,
    unpack_svndiff_instruction,
    DELTA_WINDOW_SIZE,
    MAX_ENCODED_INT_LEN,
    SVNDIFF0_HEADER,
    TXDELTA_INVALID,
    TXDELTA_NEW,
    TXDELTA_SOURCE,
    TXDELTA_TARGET,
)
from subvertpy.tests import TestCase


class DeltaTests(TestCase):
    def setUp(self):
        super(DeltaTests, self).setUp()
        self.windows = []

    def storing_window_handler(self, window):
        self.windows.append(window)

    def test_send_stream(self):
        stream = BytesIO(b"foo")
        send_stream(stream, self.storing_window_handler)
        self.assertEqual([(0, 0, 3, 0, [(2, 0, 3)], b"foo"), None], self.windows)

    def test_send_stream_returns_md5(self):
        import hashlib

        data = b"hello world"
        stream = BytesIO(data)
        result = send_stream(stream, self.storing_window_handler)
        self.assertEqual(hashlib.md5(data).digest(), result)

    def test_send_stream_empty(self):
        stream = BytesIO(b"")
        send_stream(stream, self.storing_window_handler)
        self.assertEqual([None], self.windows)

    def test_send_stream_non_bytes_raises(self):
        import io

        stream = io.StringIO("text")
        self.assertRaises(TypeError, send_stream, stream, self.storing_window_handler)

    def test_send_stream_custom_block_size(self):
        data = b"abcdef"
        stream = BytesIO(data)
        send_stream(stream, self.storing_window_handler, block_size=3)
        # Should get two windows plus None
        self.assertEqual(3, len(self.windows))
        self.assertIsNone(self.windows[-1])

    def test_apply_delta(self):
        stream = BytesIO()
        source = b"(source)"
        handler = apply_txdelta_handler(source, stream)

        new = b"(new)"
        ops = (  # (action, offset, length)
            (TXDELTA_NEW, 0, len(new)),
            (TXDELTA_SOURCE, 0, len(source)),
            (TXDELTA_TARGET, len(new), len(b"(s")),  # Copy "(s"
            (TXDELTA_TARGET, len(b"(n"), len(b"ew)")),  # Copy "ew)"
            # Copy as target is generated
            (TXDELTA_TARGET, len(new + source), len(b"(sew)") * 2),
        )
        result = b"(new)(source)(sew)(sew)(sew)"

        # (source offset, source length, result length, src_ops, ops, new)
        handler((0, len(source), len(result), 0, ops, new))
        handler(None)
        self.assertEqual(result, stream.getvalue())

    def test_apply_txdelta_handler_none_is_noop(self):
        stream = BytesIO()
        handler = apply_txdelta_handler(b"", stream)
        handler(None)
        self.assertEqual(b"", stream.getvalue())

    def test_apply_txdelta_handler_chunks(self):
        source_chunks = [b"hello", b" world"]
        target_chunks = []
        handler = apply_txdelta_handler_chunks(source_chunks, target_chunks)
        new_data = b"replaced"
        ops = [(TXDELTA_NEW, 0, len(new_data))]
        handler((0, 11, len(new_data), 0, ops, new_data))
        handler(None)
        self.assertEqual([bytearray(b"replaced")], target_chunks)

    def test_apply_txdelta_handler_chunks_source_copy(self):
        source_chunks = [b"hello"]
        target_chunks = []
        handler = apply_txdelta_handler_chunks(source_chunks, target_chunks)
        ops = [(TXDELTA_SOURCE, 0, 5)]
        handler((0, 5, 5, 0, ops, b""))
        handler(None)
        self.assertEqual([bytearray(b"hello")], target_chunks)


class TxDeltaApplyOpsTests(TestCase):
    def test_source_copy(self):
        result = txdelta_apply_ops(0, [(TXDELTA_SOURCE, 0, 3)], b"", b"abc")
        self.assertEqual(bytearray(b"abc"), result)

    def test_new_data(self):
        result = txdelta_apply_ops(0, [(TXDELTA_NEW, 0, 3)], b"xyz", b"")
        self.assertEqual(bytearray(b"xyz"), result)

    def test_target_copy(self):
        # First add some new data, then copy from target
        ops = [
            (TXDELTA_NEW, 0, 2),
            (TXDELTA_TARGET, 0, 4),
        ]
        result = txdelta_apply_ops(0, ops, b"ab", b"")
        self.assertEqual(bytearray(b"ababab"), result)

    def test_invalid_action_raises(self):
        self.assertRaises(
            Exception, txdelta_apply_ops, 0, [(TXDELTA_INVALID, 0, 1)], b"", b"x"
        )

    def test_empty_ops(self):
        result = txdelta_apply_ops(0, [], b"", b"source")
        self.assertEqual(bytearray(b""), result)


class ApplyTxDeltaWindowTests(TestCase):
    def test_simple_new(self):
        window = (0, 0, 5, 0, [(TXDELTA_NEW, 0, 5)], b"hello")
        result = apply_txdelta_window(b"", window)
        self.assertEqual(bytearray(b"hello"), result)

    def test_source_copy(self):
        window = (2, 3, 3, 0, [(TXDELTA_SOURCE, 0, 3)], b"")
        result = apply_txdelta_window(b"XXabcXX", window)
        self.assertEqual(bytearray(b"abc"), result)

    def test_tview_len_mismatch_raises(self):
        # tview_len (10) doesn't match actual result length (5)
        window = (0, 0, 10, 0, [(TXDELTA_NEW, 0, 5)], b"hello")
        self.assertRaises(AssertionError, apply_txdelta_window, b"", window)


class MarshallTests(TestCase):
    def test_encode_length(self):
        self.assertEqual(bytearray(b"\x81\x02"), encode_length(130))

    def test_encode_length_zero(self):
        self.assertEqual(bytearray(b"\x00"), encode_length(0))

    def test_encode_length_small(self):
        self.assertEqual(bytearray(b"\x01"), encode_length(1))
        self.assertEqual(bytearray(b"\x7f"), encode_length(127))

    def test_encode_length_large(self):
        encoded = encode_length(16384)
        decoded, _ = decode_length(encoded)
        self.assertEqual(16384, decoded)

    def test_roundtrip_length(self):
        self.assertEqual((42, bytes()), decode_length(encode_length(42)))

    def test_roundtrip_length_various(self):
        for val in [0, 1, 127, 128, 255, 256, 1000, 65535, 100000]:
            decoded, remainder = decode_length(encode_length(val))
            self.assertEqual(val, decoded)
            self.assertEqual(bytes(), remainder)

    def test_roundtrip_window(self):
        mywindow = (0, 0, 3, 1, [(2, 0, 3)], b"foo")
        self.assertEqual([mywindow], list(unpack_svndiff0(pack_svndiff0([mywindow]))))

    def test_roundtrip_window_multiple(self):
        w1 = (0, 0, 3, 1, [(TXDELTA_NEW, 0, 3)], b"foo")
        w2 = (0, 0, 3, 1, [(TXDELTA_NEW, 0, 3)], b"bar")
        result = list(unpack_svndiff0(pack_svndiff0([w1, w2])))
        self.assertEqual([w1, w2], result)

    def test_pack_svndiff0_header(self):
        packed = pack_svndiff0([])
        self.assertTrue(packed.startswith(SVNDIFF0_HEADER))

    def test_unpack_svndiff0_empty(self):
        result = list(unpack_svndiff0(SVNDIFF0_HEADER))
        self.assertEqual([], result)


class SvndiffInstructionTests(TestCase):
    def test_pack_unpack_new_short(self):
        instr = (TXDELTA_NEW, 0, 10)
        packed = pack_svndiff_instruction(instr)
        unpacked, remainder = unpack_svndiff_instruction(packed)
        self.assertEqual(instr, unpacked)
        self.assertEqual(b"", remainder)

    def test_pack_unpack_source(self):
        instr = (TXDELTA_SOURCE, 5, 10)
        packed = pack_svndiff_instruction(instr)
        unpacked, remainder = unpack_svndiff_instruction(packed)
        self.assertEqual(instr, unpacked)
        self.assertEqual(b"", remainder)

    def test_pack_unpack_target(self):
        instr = (TXDELTA_TARGET, 3, 7)
        packed = pack_svndiff_instruction(instr)
        unpacked, remainder = unpack_svndiff_instruction(packed)
        self.assertEqual(instr, unpacked)
        self.assertEqual(b"", remainder)

    def test_pack_unpack_long_length(self):
        # Length >= 0x3f triggers the longer encoding path
        instr = (TXDELTA_NEW, 0, 100)
        packed = pack_svndiff_instruction(instr)
        unpacked, remainder = unpack_svndiff_instruction(packed)
        self.assertEqual(instr, unpacked)
        self.assertEqual(b"", remainder)

    def test_pack_unpack_source_long(self):
        instr = (TXDELTA_SOURCE, 200, 100)
        packed = pack_svndiff_instruction(instr)
        unpacked, _ = unpack_svndiff_instruction(packed)
        self.assertEqual(instr, unpacked)

    def test_roundtrip_through_window(self):
        """Pack instructions into a window, then unpack."""
        ops = [
            (TXDELTA_NEW, 0, 5),
            (TXDELTA_SOURCE, 0, 3),
        ]
        new_data = b"hello"
        window = (0, 3, 8, len(ops), ops, new_data)
        packed = pack_svndiff0_window(window)
        # Pack into full svndiff and unpack
        full = SVNDIFF0_HEADER + bytes(packed)
        result = list(unpack_svndiff0(full))
        self.assertEqual(1, len(result))
        self.assertEqual(window, result[0])


class ConstantsTests(TestCase):
    def test_txdelta_constants(self):
        self.assertEqual(TXDELTA_SOURCE, 0)
        self.assertEqual(TXDELTA_TARGET, 1)
        self.assertEqual(TXDELTA_NEW, 2)
        self.assertEqual(TXDELTA_INVALID, 3)

    def test_max_encoded_int_len(self):
        self.assertEqual(MAX_ENCODED_INT_LEN, 10)

    def test_delta_window_size(self):
        self.assertEqual(DELTA_WINDOW_SIZE, 102400)

    def test_svndiff0_header(self):
        self.assertEqual(SVNDIFF0_HEADER, b"SVN\0")



#!/usr/bin/python

import cmd
import subvertpy
from subvertpy.ra import RemoteAccess
import sys

if len(sys.argv) == 1:
    print("Usage: %s <url>" % sys.argv)

url = sys.argv[1]

conn = RemoteAccess(url)


def log_printer(changed_paths, rev, revprops, has_children=None):
    print("=" * 79)
    print("%d:" % rev)
    print("Revision properties:")
    for entry in revprops.items():
        print("  %s: %s" % entry)
    print("")

    if changed_paths is None:
        return
    print("Changed paths:")
    for path, (action, from_path, from_rev) in changed_paths.items():
        print("  %s (%s)" % (path, action))


class RaCmd(cmd.Cmd):
    @staticmethod
    def parse_path_revnum(line):
        args = line.split(" ")
        if len(args) == 0:
            return ".", -1
        elif len(args) == 1:
            return args[0], -1
        elif len(args) == 2:
            return args[0], int(args[1])
        else:
            raise Exception("Too much arguments (%r), expected 2" % (args,))

    def do_help(self, args):
        for name in sorted(self.__class__.__dict__):
            if name.startswith("do_"):
                print(name[3:])

    def do_stat(self, args):
        path, revnum = self.parse_path_revnum(args)
        print(conn.stat(path, revnum))

    def do_ls(self, args):
        path, revnum = self.parse_path_revnum(args)
        (dirents, fetched_rev, props) = conn.get_dir(path, revnum)
        for name in dirents:
            print(name)

    def do_cat(self, args):
        path, revnum = self.parse_path_revnum(args)
        outf = getattr(sys.stdout, "buffer", sys.stdout)
        (fetched_rev, props) = conn.get_file(path, outf, revnum)

    def do_reparent(self, args):
        conn.reparent(args)

    def do_set_revprop(self, args):
        (revnum, name, value) = args.split(" ", 2)
        conn.change_rev_prop(int(revnum), name, value)

    def do_has_capability(self, args):
        print(conn.has_capability(args))

    def do_revprops(self, args):
        for item in conn.rev_proplist(int(args)).items():
            print("%s: %s" % item)

    def do_check_path(self, args):
        path, revnum = self.parse_path_revnum(args)
        kind = conn.check_path(path, revnum)
        if kind == subvertpy.NODE_DIR:
            print("dir")
        elif kind == subvertpy.NODE_FILE:
            print("file")
        else:
            print("nonexistant")

    def do_uuid(self, args):
        print(conn.get_uuid())

    def do_get_repos_root(self, args):
        print(conn.get_repos_root())

    def do_log(self, args):
        conn.get_log(
            callback=log_printer,
            paths=None,
            start=0,
            end=conn.get_latest_revnum(),
            discover_changed_paths=True,
        )


cmdline = RaCmd()
cmdline.cmdloop()



# Copyright (C) 2005-2007 Jelmer Vernooij <jelmer@jelmer.uk>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Subversion ra library tests."""

from io import BytesIO

from subvertpy import (
    NODE_DIR,
    NODE_NONE,
    SubversionException,
    ra,
)
from subvertpy.tests import (
    SubversionTestCase,
    TestCase,
)


class VersionTest(TestCase):
    def test_version_length(self):
        self.assertEqual(4, len(ra.version()))

    def test_api_version_length(self):
        self.assertEqual(4, len(ra.api_version()))

    def test_api_version_later_same(self):
        self.assertTrue(ra.api_version() <= ra.version())


class TestRemoteAccessUnknown(TestCase):
    def test_unknown_url(self):
        self.assertRaises(SubversionException, ra.RemoteAccess, "bla://")

    def test_unknown_url_bytes(self):
        self.assertRaises(SubversionException, ra.RemoteAccess, b"bla://")

    def test_url_handlers_populated(self):
        self.assertIn("svn", ra.url_handlers)
        self.assertIn("svn+ssh", ra.url_handlers)
        self.assertIn("http", ra.url_handlers)
        self.assertIn("https", ra.url_handlers)
        self.assertIn("file", ra.url_handlers)


class TestRemoteAccess(SubversionTestCase):
    def setUp(self):
        super(TestRemoteAccess, self).setUp()
        self.repos_url = self.make_repository("d")
        self.ra = ra.RemoteAccess(
            self.repos_url, auth=ra.Auth([ra.get_username_provider()])
        )

    def tearDown(self):
        del self.ra
        super(TestRemoteAccess, self).tearDown()

    def commit_editor(self):
        return self.get_commit_editor(self.repos_url)

    def do_commit(self):
        dc = self.get_commit_editor(self.repos_url)
        dc.add_dir("foo")
        dc.close()

    def test_repr(self):
        self.assertEqual('RemoteAccess("%s")' % self.repos_url, repr(self.ra))

    def test_latest_revnum(self):
        self.assertEqual(0, self.ra.get_latest_revnum())

    def test_latest_revnum_one(self):
        self.do_commit()
        self.assertEqual(1, self.ra.get_latest_revnum())

    def test_get_uuid(self):
        self.assertEqual(36, len(self.ra.get_uuid()))

    def test_get_repos_root(self):
        self.assertEqual(self.repos_url, self.ra.get_repos_root())

    def test_get_url(self):
        self.assertEqual(self.repos_url, self.ra.get_session_url())

    def test_reparent(self):
        self.ra.reparent(self.repos_url)

    def test_has_capability(self):
        self.assertRaises(SubversionException, self.ra.has_capability, "FOO")

    def test_get_dir(self):
        ret = self.ra.get_dir("", 0)
        self.assertIsInstance(ret, tuple)

    def test_get_dir_leading_slash(self):
        ret = self.ra.get_dir("/", 0)
        self.assertIsInstance(ret, tuple)

    def test_get_dir_kind(self):
        self.do_commit()
        (dirents, fetch_rev, props) = self.ra.get_dir("/", 1, fields=ra.DIRENT_KIND)
        self.assertIsInstance(props, dict)
        self.assertEqual(1, fetch_rev)
        self.assertEqual(NODE_DIR, dirents["foo"]["kind"])

    def test_change_rev_prop(self):
        self.do_commit()
        self.ra.change_rev_prop(1, "foo", "bar")

    def test_rev_proplist(self):
        self.assertIsInstance(self.ra.rev_proplist(0), dict)

    def test_do_diff(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra.do_diff(1, "", self.ra.get_repos_root(), MyEditor())
        reporter.set_path("", 0, True)
        reporter.finish()
        self.assertRaises(RuntimeError, reporter.finish)
        self.assertRaises(RuntimeError, reporter.set_path, "", 0, True)

    def test_iter_log_invalid(self):
        self.assertRaises(
            SubversionException,
            list,
            self.ra.iter_log(
                ["idontexist"], 0, 0, revprops=["svn:date", "svn:author", "svn:log"]
            ),
        )
        self.assertRaises(
            SubversionException,
            list,
            self.ra.iter_log(
                [""], 0, 1000, revprops=["svn:date", "svn:author", "svn:log"]
            ),
        )

    def test_iter_log(self):
        def check_results(returned):
            self.assertEqual(2, len(returned))
            self.assertTrue(len(returned[0]) in (3, 4))
            if len(returned[0]) == 3:
                (paths, revnum, props) = returned[0]
            else:
                (paths, revnum, props, has_children) = returned[0]
            self.assertEqual(None, paths)
            self.assertEqual(revnum, 0)
            self.assertEqual(["svn:date"], list(props.keys()))
            if len(returned[1]) == 3:
                (paths, revnum, props) = returned[1]
            else:
                (paths, revnum, props, has_children) = returned[1]
            self.assertEqual({"/foo": ("A", None, -1, NODE_DIR)}, paths)
            self.assertEqual(revnum, 1)
            self.assertEqual(
                set(["svn:date", "svn:author", "svn:log"]), set(props.keys())
            )

        returned = list(
            self.ra.iter_log([""], 0, 0, revprops=["svn:date", "svn:author", "svn:log"])
        )
        self.assertEqual(1, len(returned))
        self.do_commit()
        returned = list(
            self.ra.iter_log(
                None,
                0,
                1,
                discover_changed_paths=True,
                strict_node_history=False,
                revprops=["svn:date", "svn:author", "svn:log"],
            )
        )
        check_results(returned)

    def test_get_log(self):
        returned = []

        def cb(*args):
            returned.append(args)

        def check_results(returned):
            self.assertEqual(2, len(returned))
            self.assertTrue(len(returned[0]) in (3, 4))
            if len(returned[0]) == 3:
                (paths, revnum, props) = returned[0]
            else:
                (paths, revnum, props, has_children) = returned[0]
            self.assertEqual(None, paths)
            self.assertEqual(revnum, 0)
            self.assertEqual(["svn:date"], list(props.keys()))
            if len(returned[1]) == 3:
                (paths, revnum, props) = returned[1]
            else:
                (paths, revnum, props, has_children) = returned[1]
            self.assertEqual({"/foo": ("A", None, -1)}, paths)
            self.assertEqual(revnum, 1)
            self.assertEqual(
                set(["svn:date", "svn:author", "svn:log"]), set(props.keys())
            )

        self.ra.get_log(cb, [""], 0, 0, revprops=["svn:date", "svn:author", "svn:log"])
        self.assertEqual(1, len(returned))
        self.do_commit()
        returned = []
        self.ra.get_log(
            cb,
            None,
            0,
            1,
            discover_changed_paths=True,
            strict_node_history=False,
            revprops=["svn:date", "svn:author", "svn:log"],
        )
        check_results(returned)

    def test_get_log_cancel(self):

        def cb(*args):
            raise KeyError

        self.do_commit()
        self.assertRaises(
            KeyError,
            self.ra.get_log,
            cb,
            [""],
            0,
            0,
            revprops=["svn:date", "svn:author", "svn:log"],
        )

    def test_get_commit_editor_double_close(self):
        def mycb(*args):
            pass

        editor = self.ra.get_commit_editor({"svn:log": "foo"}, mycb)
        dir = editor.open_root()
        dir.close()
        self.assertRaises(RuntimeError, dir.close)
        editor.close()
        self.assertRaises(RuntimeError, editor.close)
        self.assertRaises(RuntimeError, editor.abort)

    def test_get_commit_editor_busy(self):
        def mycb(rev):
            pass

        editor = self.ra.get_commit_editor({"svn:log": "foo"}, mycb)
        self.assertRaises(
            ra.BusyException, self.ra.get_commit_editor, {"svn:log": "foo"}, mycb
        )
        editor.abort()

    def test_get_commit_editor_double_open(self):
        def mycb(rev):
            pass

        editor = self.ra.get_commit_editor({"svn:log": "foo"}, mycb)
        root = editor.open_root()
        root.add_directory("somedir")
        self.assertRaises(RuntimeError, root.add_directory, "foo")

    def test_get_commit_editor_custom_revprops(self):
        if ra.version()[:2] < (1, 5):
            return

        def mycb(paths, rev, revprops):
            pass

        editor = self.ra.get_commit_editor(
            {"svn:log": "foo", "bar:foo": "bla", "svn:custom:blie": "bloe"}, mycb
        )
        root = editor.open_root()
        root.add_directory("somedir").close()
        root.close()
        editor.close()

        revprops = self.ra.rev_proplist(1)
        self.assertEqual(
            set(["bar:foo", "svn:author", "svn:custom:blie", "svn:date", "svn:log"]),
            set(revprops.keys()),
            "result: %r" % revprops,
        )

    def test_get_commit_editor_context_manager(self):
        def mycb(paths, rev, revprops):
            pass

        editor = self.ra.get_commit_editor({"svn:log": "foo"}, mycb)
        self.assertIs(editor, editor.__enter__())
        dir = editor.open_root(0)
        subdir = dir.add_directory("foo")
        self.assertIs(subdir, subdir.__enter__())
        subdir.__exit__(None, None, None)
        dir.__exit__(None, None, None)
        editor.__exit__(None, None, None)

    def test_get_commit_editor(self):
        def mycb(paths, rev, revprops):
            pass

        editor = self.ra.get_commit_editor({"svn:log": "foo"}, mycb)
        dir = editor.open_root(0)
        subdir = dir.add_directory("foo")
        subdir.close()
        dir.close()
        editor.close()

    def test_get_commit_editor_with_lock_tokens(self):
        def mycb(paths, rev, revprops):
            pass

        editor = self.ra.get_commit_editor(
            {"svn:log": "with locks"}, mycb, lock_tokens={}, keep_locks=True
        )
        root = editor.open_root()
        root.close()
        editor.close()

    def test_get_log_with_limit(self):
        self.do_commit()
        returned = []

        def cb(*args):
            returned.append(args)

        self.ra.get_log(cb, [""], 0, 1, limit=1, include_merged_revisions=False)
        self.assertEqual(1, len(returned))

    def test_change_rev_prop_old_value(self):
        self.do_commit()
        old = self.ra.rev_proplist(1).get("foo")
        self.ra.change_rev_prop(1, "foo", "bar", old_value=old)
        self.assertEqual(b"bar", self.ra.rev_proplist(1)["foo"])

    def test_do_diff_with_options(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra.do_diff(
            1,
            "",
            self.ra.get_repos_root(),
            MyEditor(),
            recurse=True,
            ignore_ancestry=True,
            text_deltas=True,
        )
        reporter.set_path("", 0, True)
        reporter.finish()

    def test_replay_send_deltas(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def apply_textdelta(self, base_checksum):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        self.ra.replay(1, 0, MyEditor(), send_deltas=False)

    def test_replay_range_send_deltas(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def apply_textdelta(self, base_checksum):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        editors = []

        def revstart_cb(rev, revprops):
            e = MyEditor()
            editors.append(e)
            return e

        def revfinish_cb(rev, revprops, editor):
            pass

        self.ra.replay_range(1, 1, 0, (revstart_cb, revfinish_cb), send_deltas=False)
        self.assertEqual(1, len(editors))

    def test_do_update_with_options(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra.do_update(
            1, "", True, MyEditor(), send_copyfrom_args=True, ignore_ancestry=True
        )
        reporter.set_path("", 0, True)
        reporter.finish()

    def test_do_switch_with_options(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra.do_switch(
            1,
            "",
            True,
            self.repos_url,
            MyEditor(),
            send_copyfrom_args=True,
            ignore_ancestry=True,
        )
        reporter.set_path("", 0, True)
        reporter.finish()

    def test_get_file_revs_include_merged(self):
        dc = self.commit_editor()
        f = dc.add_file("filerevs")
        f.modify(b"v1")
        dc.close()
        revs = []

        def handler(path, rev, rev_props, prop_diffs=None):
            revs.append(rev)

        self.ra.get_file_revs("filerevs", 0, 1, handler, include_merged_revisions=True)
        self.assertTrue(len(revs) > 0)

    def test_mergeinfo_with_options(self):
        self.do_commit()
        result = self.ra.mergeinfo(
            [""], revision=1, inherit=ra.MERGEINFO_INHERITED, include_descendants=True
        )
        if result is not None:
            self.assertIsInstance(result, dict)

    def test_commit_file_props(self):
        cb = self.commit_editor()
        f = cb.add_file("bar")
        f.modify(b"a")
        f.change_prop("bla:bar", "blie")
        cb.close()

        cb = self.commit_editor()
        f = cb.open_file("bar")
        f.change_prop("bla:bar", None)
        cb.close()

        stream = BytesIO()
        props = self.ra.get_file("bar", stream, 1)[1]
        self.assertEqual(b"blie", props.get("bla:bar"))
        stream = BytesIO()
        props = self.ra.get_file("bar", stream, 2)[1]
        self.assertIs(None, props.get("bla:bar"))

    def test_get_file_revs(self):
        cb = self.commit_editor()
        cb.add_file("bar").modify(b"a")
        cb.close()

        cb = self.commit_editor()
        f = cb.open_file("bar")
        f.modify(b"b")
        f.change_prop("bla", "bloe")
        cb.close()

        rets = []

        def handle(path, rev, props, from_merge=None):
            rets.append((path, rev, props))

        self.ra.get_file_revs("bar", 1, 2, handle)

        self.assertEqual(2, len(rets))
        self.assertEqual(1, rets[0][1])
        self.assertEqual(2, rets[1][1])
        self.assertEqual("/bar", rets[0][0])
        self.assertEqual("/bar", rets[1][0])

    def test_get_file(self):
        cb = self.commit_editor()
        cb.add_file("bar").modify(b"a")
        cb.close()

        stream = BytesIO()
        self.ra.get_file("bar", stream, 1)
        stream.seek(0)
        self.assertEqual(b"a", stream.read())

        stream = BytesIO()
        self.ra.get_file("/bar", stream, 1)
        stream.seek(0)
        self.assertEqual(b"a", stream.read())

    def test_get_locations_root(self):
        self.assertEqual({0: "/"}, self.ra.get_locations("", 0, [0]))

    def test_check_path(self):
        cb = self.commit_editor()
        cb.add_dir("bar")
        cb.close()

        self.assertEqual(NODE_DIR, self.ra.check_path("bar", 1))
        self.assertEqual(NODE_DIR, self.ra.check_path("bar/", 1))
        self.assertEqual(NODE_NONE, self.ra.check_path("blaaaa", 1))

    def test_stat(self):
        cb = self.commit_editor()
        cb.add_dir("bar")
        cb.close()

        ret = self.ra.stat("bar", 1)
        self.assertEqual(
            set(["last_author", "kind", "created_rev", "has_props", "time", "size"]),
            set(ret.keys()),
        )

    def test_get_locations_dir(self):
        cb = self.commit_editor()
        cb.add_dir("bar")
        cb.close()

        cb = self.commit_editor()
        cb.add_dir("bla", "bar", 1)
        cb.close()

        cb = self.commit_editor()
        cb.delete("bar")
        cb.close()

        self.assertEqual(
            {1: "/bar", 2: "/bla"}, self.ra.get_locations("bla", 2, [1, 2])
        )

        self.assertEqual(
            {1: "/bar", 2: "/bar"}, self.ra.get_locations("bar", 1, [1, 2])
        )

        self.assertEqual(
            {1: "/bar", 2: "/bar"}, self.ra.get_locations("bar", 2, [1, 2])
        )

        self.assertEqual(
            {1: "/bar", 2: "/bla", 3: "/bla"},
            self.ra.get_locations("bla", 3, [1, 2, 3]),
        )


class TestEditorOperations(SubversionTestCase):
    """Tests for editor operations: delete_entry, open_directory, etc."""

    def setUp(self):
        super(TestEditorOperations, self).setUp()
        self.repos_url = self.make_repository("d")

    def commit_editor(self):
        return self.get_commit_editor(self.repos_url)

    def test_delete_entry(self):
        dc = self.commit_editor()
        dc.add_file("todelete").modify(b"bye")
        dc.close()

        dc = self.commit_editor()
        dc.delete("todelete")
        dc.close()

        r = ra.RemoteAccess(self.repos_url, auth=ra.Auth([ra.get_username_provider()]))
        self.assertEqual(NODE_NONE, r.check_path("todelete", 2))

    def test_open_directory(self):
        dc = self.commit_editor()
        subdir = dc.add_dir("mydir")
        subdir.add_file("mydir/inner").modify(b"data")
        dc.close()

        dc = self.commit_editor()
        subdir = dc.open_dir("mydir")
        subdir.add_file("mydir/another").modify(b"more")
        dc.close()

        r = ra.RemoteAccess(self.repos_url, auth=ra.Auth([ra.get_username_provider()]))
        stream = BytesIO()
        r.get_file("mydir/another", stream, 2)
        stream.seek(0)
        self.assertEqual(b"more", stream.read())

    def test_dir_change_prop(self):
        dc = self.commit_editor()
        subdir = dc.add_dir("propdir")
        subdir.change_prop("myprop", "myval")
        dc.close()

        r = ra.RemoteAccess(self.repos_url, auth=ra.Auth([ra.get_username_provider()]))
        (dirents, rev, props) = r.get_dir("propdir", 1)
        self.assertIn("myprop", props)
        self.assertEqual(b"myval", props["myprop"])

    def test_absent_file(self):
        # absent_file is used by editors to signal a file is not present
        # We test it via the low-level commit editor
        r = ra.RemoteAccess(self.repos_url, auth=ra.Auth([ra.get_username_provider()]))
        editor = r.get_commit_editor({"svn:log": "absent test"})
        root = editor.open_root()
        root.absent_file("ghost")
        root.close()
        editor.close()

    def test_absent_directory(self):
        r = ra.RemoteAccess(self.repos_url, auth=ra.Auth([ra.get_username_provider()]))
        editor = r.get_commit_editor({"svn:log": "absent dir test"})
        root = editor.open_root()
        root.absent_directory("ghostdir")
        root.close()
        editor.close()


class AuthTests(TestCase):
    def test_not_list(self):
        self.assertRaises(TypeError, ra.Auth, ra.get_simple_provider())

    def test_not_registered(self):
        auth = ra.Auth([])
        self.assertRaises(
            SubversionException, auth.credentials, "svn.simple", "MyRealm"
        )

    def test_simple(self):
        auth = ra.Auth(
            [
                ra.get_simple_prompt_provider(
                    lambda realm, uname, may_save: ("foo", "geheim", False), 0
                )
            ]
        )
        creds = auth.credentials("svn.simple", "MyRealm")
        self.assertEqual(("foo", "geheim", 0), next(creds))
        self.assertRaises(StopIteration, next, creds)

    def test_username(self):
        auth = ra.Auth(
            [
                ra.get_username_prompt_provider(
                    lambda realm, may_save: ("somebody", False), 0
                )
            ]
        )
        creds = auth.credentials("svn.username", "MyRealm")
        self.assertEqual(("somebody", 0), next(creds))
        self.assertRaises(StopIteration, next, creds)

    def test_client_cert(self):
        auth = ra.Auth(
            [
                ra.get_ssl_client_cert_prompt_provider(
                    lambda realm, may_save: ("filename", False), 0
                )
            ]
        )
        creds = auth.credentials("svn.ssl.client-cert", "MyRealm")
        self.assertEqual(("filename", False), next(creds))
        self.assertRaises(StopIteration, next, creds)

    def test_client_cert_pw(self):
        auth = ra.Auth(
            [
                ra.get_ssl_client_cert_pw_prompt_provider(
                    lambda realm, may_save: ("supergeheim", False), 0
                )
            ]
        )
        creds = auth.credentials("svn.ssl.client-passphrase", "MyRealm")
        self.assertEqual(("supergeheim", False), next(creds))
        self.assertRaises(StopIteration, next, creds)

    def test_server_trust(self):
        auth = ra.Auth(
            [
                ra.get_ssl_server_trust_prompt_provider(
                    lambda realm, failures, certinfo, may_save: (42, False)
                )
            ]
        )
        auth.set_parameter("svn:auth:ssl:failures", 23)
        creds = auth.credentials("svn.ssl.server", "MyRealm")
        self.assertEqual((42, 0), next(creds))
        self.assertRaises(StopIteration, next, creds)

    def test_server_untrust(self):
        auth = ra.Auth(
            [
                ra.get_ssl_server_trust_prompt_provider(
                    lambda realm, failures, certinfo, may_save: None
                )
            ]
        )
        auth.set_parameter("svn:auth:ssl:failures", 23)
        creds = auth.credentials("svn.ssl.server", "MyRealm")
        self.assertRaises(StopIteration, next, creds)

    def test_retry(self):
        self.i = 0

        def inc_foo(realm, may_save):
            self.i += 1
            return ("somebody%d" % self.i, False)

        auth = ra.Auth([ra.get_username_prompt_provider(inc_foo, 2)])
        creds = auth.credentials("svn.username", "MyRealm")
        self.assertEqual(("somebody1", 0), next(creds))
        self.assertEqual(("somebody2", 0), next(creds))
        self.assertEqual(("somebody3", 0), next(creds))
        self.assertRaises(StopIteration, next, creds)

    def test_set_default_username(self):
        a = ra.Auth([])
        a.set_parameter("svn:auth:username", "foo")
        self.assertEqual("foo", a.get_parameter("svn:auth:username"))

    def test_set_default_password(self):
        a = ra.Auth([])
        a.set_parameter("svn:auth:password", "bar")
        self.assertEqual("bar", a.get_parameter("svn:auth:password"))

    def test_platform_auth_providers(self):
        ra.Auth(ra.get_platform_specific_client_providers())


class TestProviders(TestCase):
    def test_get_simple_provider(self):
        provider = ra.get_simple_provider()
        self.assertIsNotNone(provider)

    def test_get_username_provider(self):
        provider = ra.get_username_provider()
        self.assertIsNotNone(provider)

    def test_get_ssl_client_cert_file_provider(self):
        provider = ra.get_ssl_client_cert_file_provider()
        self.assertIsNotNone(provider)

    def test_get_ssl_client_cert_pw_file_provider(self):
        provider = ra.get_ssl_client_cert_pw_file_provider()
        self.assertIsNotNone(provider)

    def test_get_ssl_server_trust_file_provider(self):
        provider = ra.get_ssl_server_trust_file_provider()
        self.assertIsNotNone(provider)

    def test_print_modules(self):
        result = ra.print_modules()
        self.assertIsInstance(result, bytes)
        self.assertIn(b"ra_local", result)


class TestRemoteAccessProperties(SubversionTestCase):
    def setUp(self):
        super(TestRemoteAccessProperties, self).setUp()
        self.repos_url = self.make_repository("d")
        self.ra_ctx = ra.RemoteAccess(
            self.repos_url, auth=ra.Auth([ra.get_username_provider()])
        )

    def tearDown(self):
        del self.ra_ctx
        super(TestRemoteAccessProperties, self).tearDown()

    def test_url_property(self):
        self.assertEqual(self.repos_url, self.ra_ctx.url)

    def test_busy_property(self):
        self.assertFalse(self.ra_ctx.busy)

    def test_get_lock_nonexistent(self):
        self.do_commit()
        lock = self.ra_ctx.get_lock("/nonexistent")
        self.assertIsNone(lock)

    def test_has_capability_mergeinfo(self):
        result = self.ra_ctx.has_capability("mergeinfo")
        self.assertIsInstance(result, bool)

    def test_has_capability_unknown(self):
        self.assertRaises(SubversionException, self.ra_ctx.has_capability, "bogus")

    def do_commit(self):
        dc = self.get_commit_editor(self.repos_url)
        dc.add_dir("foo")
        dc.close()

    def test_do_update(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra_ctx.do_update(1, "", True, MyEditor())
        reporter.set_path("", 0, True)
        reporter.finish()

    def test_replay(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        self.ra_ctx.replay(1, 0, MyEditor())

    def test_get_location_segments(self):
        self.do_commit()
        segments = []

        def rcvr(range_start, range_end, path):
            segments.append((range_start, range_end, path))

        self.ra_ctx.get_location_segments("foo", 1, 1, 0, rcvr)
        self.assertEqual(1, len(segments))
        self.assertEqual("foo", segments[0][2])

    def test_get_commit_editor_abort(self):
        def mycb(*args):
            pass

        editor = self.ra_ctx.get_commit_editor({"svn:log": "test"}, mycb)
        editor.abort()

    def test_editor_dir_context_manager(self):
        def mycb(*args):
            pass

        editor = self.ra_ctx.get_commit_editor({"svn:log": "test"}, mycb)
        with editor:
            root = editor.open_root()
            with root:
                subdir = root.add_directory("mydir")
                with subdir:
                    pass

    def test_get_locks_empty(self):
        self.do_commit()
        locks = self.ra_ctx.get_locks("")
        self.assertIsInstance(locks, dict)
        self.assertEqual({}, locks)

    def test_lock_unlock(self):
        # Create a file to lock
        dc = self.get_commit_editor(self.repos_url)
        f = dc.add_file("lockme")
        f.modify(b"content")
        dc.close()

        lock_results = []

        def lock_cb(path, do_lock, lock, ra_err):
            lock_results.append((path, do_lock, lock, ra_err))

        self.ra_ctx.lock({b"lockme": 1}, "locking", False, lock_cb)
        self.assertEqual(1, len(lock_results))

        # Verify lock exists
        locks = self.ra_ctx.get_locks("")
        self.assertIn("/lockme", locks)

        # Lock is a tuple: (path, token, owner, comment, is_dav_comment,
        #                    creation_date, expiration_date)
        lock_tuple = locks["/lockme"]
        token = lock_tuple[1]

        # Now unlock
        unlock_results = []

        def unlock_cb(path, do_lock, lock, ra_err):
            unlock_results.append((path, do_lock, lock, ra_err))

        self.ra_ctx.unlock({b"lockme": token}, False, unlock_cb)
        self.assertEqual(1, len(unlock_results))

        # Verify lock is gone
        locks = self.ra_ctx.get_locks("")
        self.assertEqual({}, locks)

    def test_get_locks_with_lock(self):
        dc = self.get_commit_editor(self.repos_url)
        f = dc.add_file("myfile")
        f.modify(b"data")
        dc.close()

        def lock_cb(path, do_lock, lock, ra_err):
            pass

        self.ra_ctx.lock({b"myfile": 1}, "test lock", False, lock_cb)
        locks = self.ra_ctx.get_locks("")
        self.assertIn("/myfile", locks)
        # Lock is a tuple: (path, token, owner, comment, ...)
        lock_tuple = locks["/myfile"]
        self.assertIsInstance(lock_tuple, tuple)
        self.assertEqual("/myfile", lock_tuple[0])  # path
        self.assertIsNotNone(lock_tuple[1])  # token
        self.assertEqual("test lock", lock_tuple[3])  # comment

        # cleanup
        self.ra_ctx.unlock({b"myfile": lock_tuple[1]}, False, lock_cb)

    def test_get_locks_with_depth(self):
        dc = self.get_commit_editor(self.repos_url)
        f = dc.add_file("lockdepth")
        f.modify(b"data")
        dc.close()

        def lock_cb(path, do_lock, lock, ra_err):
            pass

        self.ra_ctx.lock({b"lockdepth": 1}, "depth lock", False, lock_cb)
        locks = self.ra_ctx.get_locks("", ra.DEPTH_INFINITY)
        self.assertIn("/lockdepth", locks)

        # cleanup
        lock_tuple = locks["/lockdepth"]
        self.ra_ctx.unlock({b"lockdepth": lock_tuple[1]}, False, lock_cb)

    def test_do_switch(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def open_directory(self, *args):
                return MyDirEditor()

            def open_file(self, *args):
                return MyFileEditor()

            def delete_entry(self, *args):
                pass

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra_ctx.do_switch(1, "", True, self.repos_url, MyEditor())
        reporter.set_path("", 0, True)
        reporter.finish()

    def test_replay_range(self):
        self.do_commit()
        # Create another commit
        dc = self.get_commit_editor(self.repos_url)
        dc.add_file("bar").modify(b"content")
        dc.close()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

            def apply_textdelta(self, base_checksum=None):
                return None

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        editors = []

        def revstart_cb(rev, revprops):
            e = MyEditor()
            editors.append(e)
            return e

        def revfinish_cb(rev, revprops, editor):
            pass

        self.ra_ctx.replay_range(1, 2, 0, (revstart_cb, revfinish_cb))
        self.assertEqual(2, len(editors))

    def test_constructor_client_string_func(self):
        def client_string_func():
            return "test-client/1.0"

        ra_ctx = ra.RemoteAccess(
            self.repos_url,
            auth=ra.Auth([ra.get_username_provider()]),
            client_string_func=client_string_func,
        )
        self.assertIsNotNone(ra_ctx)
        ra_ctx.get_latest_revnum()
        del ra_ctx

    def test_constructor_uuid(self):
        ra_ctx = ra.RemoteAccess(
            self.repos_url,
            auth=ra.Auth([ra.get_username_provider()]),
            uuid=self.ra_ctx.get_uuid(),
        )
        self.assertIsNotNone(ra_ctx)
        del ra_ctx

    def test_constructor_progress_cb(self):
        progress_calls = []

        def progress_cb(progress, total):
            progress_calls.append((progress, total))

        ra_ctx = ra.RemoteAccess(
            self.repos_url,
            auth=ra.Auth([ra.get_username_provider()]),
            progress_cb=progress_cb,
        )
        ra_ctx.get_latest_revnum()
        del ra_ctx

    def test_get_simple_provider_callback(self):
        def simple_cb(realm, username, may_save):
            return ("user", "pass", False)

        provider = ra.get_simple_provider(simple_cb)
        self.assertIsNotNone(provider)

    def test_constructor_open_tmp_file_func(self):
        import tempfile

        def open_tmp_file():
            return tempfile.mktemp()

        ra_ctx = ra.RemoteAccess(
            self.repos_url,
            auth=ra.Auth([ra.get_username_provider()]),
            open_tmp_file_func=open_tmp_file,
        )
        self.assertIsNotNone(ra_ctx)
        ra_ctx.get_latest_revnum()
        del ra_ctx

    def test_progress_func(self):
        progress_calls = []

        def progress_cb(progress, total):
            progress_calls.append((progress, total))

        ra_ctx = ra.RemoteAccess(
            self.repos_url, auth=ra.Auth([ra.get_username_provider()])
        )
        ra_ctx.progress_func = progress_cb
        ra_ctx.get_latest_revnum()
        del ra_ctx

    def test_constructor_config(self):
        from subvertpy import client

        config = client.get_config()
        ra_ctx = ra.RemoteAccess(
            self.repos_url, auth=ra.Auth([ra.get_username_provider()]), config=config
        )
        self.assertIsNotNone(ra_ctx)
        ra_ctx.get_latest_revnum()
        del ra_ctx

    def test_reporter_set_path_with_depth(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def delete_entry(self, *args):
                pass

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra_ctx.do_update(1, "", True, MyEditor())
        reporter.set_path("", 0, True, None, ra.DEPTH_INFINITY)
        reporter.finish()

    def test_reporter_link_path_with_options(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def open_directory(self, *args):
                return MyDirEditor()

            def open_file(self, *args):
                return MyFileEditor()

            def delete_entry(self, *args):
                pass

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra_ctx.do_update(1, "", True, MyEditor())
        reporter.set_path("", 0, True)
        reporter.link_path(
            "foo", self.repos_url + "/foo", 1, True, None, ra.DEPTH_INFINITY
        )
        reporter.finish()

    def test_reporter_delete_path(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def delete_entry(self, *args):
                pass

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra_ctx.do_update(1, "", True, MyEditor())
        reporter.set_path("", 0, True)
        reporter.delete_path("foo")
        reporter.finish()

    def test_mergeinfo_empty(self):
        self.do_commit()
        result = self.ra_ctx.mergeinfo(["foo"])
        # No mergeinfo set, so result should be None or empty
        if result is not None:
            self.assertIsInstance(result, dict)

    def test_reporter_link_path(self):
        self.do_commit()

        class MyFileEditor:
            def change_prop(self, name, val):
                pass

            def close(self, checksum=None):
                pass

        class MyDirEditor:
            def change_prop(self, name, val):
                pass

            def add_directory(self, *args):
                return MyDirEditor()

            def open_directory(self, *args):
                return MyDirEditor()

            def add_file(self, *args):
                return MyFileEditor()

            def open_file(self, *args):
                return MyFileEditor()

            def delete_entry(self, *args):
                pass

            def close(self):
                pass

        class MyEditor:
            def set_target_revision(self, rev):
                pass

            def open_root(self, base_rev):
                return MyDirEditor()

            def close(self):
                pass

        reporter = self.ra_ctx.do_update(1, "", True, MyEditor())
        reporter.set_path("", 0, True)
        # link_path(path, url, revision, start_empty)
        reporter.link_path("foo", self.repos_url + "/foo", 1, True)
        reporter.finish()


class ConstantsTests(TestCase):
    def test_depth_constants(self):
        self.assertIsInstance(ra.DEPTH_UNKNOWN, int)
        self.assertIsInstance(ra.DEPTH_EXCLUDE, int)
        self.assertIsInstance(ra.DEPTH_EMPTY, int)
        self.assertIsInstance(ra.DEPTH_FILES, int)
        self.assertIsInstance(ra.DEPTH_IMMEDIATES, int)
        self.assertIsInstance(ra.DEPTH_INFINITY, int)

    def test_dirent_constants(self):
        self.assertIsInstance(ra.DIRENT_KIND, int)
        self.assertIsInstance(ra.DIRENT_SIZE, int)
        self.assertIsInstance(ra.DIRENT_HAS_PROPS, int)
        self.assertIsInstance(ra.DIRENT_CREATED_REV, int)
        self.assertIsInstance(ra.DIRENT_TIME, int)
        self.assertIsInstance(ra.DIRENT_LAST_AUTHOR, int)
        self.assertIsInstance(ra.DIRENT_ALL, int)

    def test_mergeinfo_constants(self):
        self.assertIsInstance(ra.MERGEINFO_EXPLICIT, int)
        self.assertIsInstance(ra.MERGEINFO_INHERITED, int)
        self.assertIsInstance(ra.MERGEINFO_NEAREST_ANCESTOR, int)

    def test_svn_revision(self):
        self.assertIsInstance(ra.SVN_REVISION, int)
